/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#ifdef __GNUC__
extern FILE *popen(const char *,const char *);
extern int pclose(FILE *);
extern int fileno(FILE *);
extern char *strdup(const char *);
#endif

#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

extern int Localmode;
int Port = 7536;
char *McastAddr = "224.0.0.75";
extern char Execname[PATH_MAX];


static double
get_double(const char *str)
{
	static const double p[] = {0.1,0.01,0.001,1e-4,1e-5,1e-6};
	double d = 0;
	char c = *str;
	double mult = 1;
	
	if (c == '-') {
		mult = -1;
		c = *++str;
	}
	while ((c >= '0') && (c <= '9')) {
		++str;
		d *= 10;
		d += (double) (c - '0');
		c = *str;
	}
	if (c == '.') {
		int comma = 0;
		c = *++str;
		while ((c >= '0') && (c <= '9')) {
			++str;
			assert(comma < (sizeof(p)/sizeof(double)));
			d += (double) (c - '0') * p[comma];
			++comma;
			c = *str;
		}
	}
	assert((c == '\t') || (c == '\n') || (c == 0));
	return d;
}


int 
setup_source(const char *IntervalArg, unsigned nouserstat, unsigned nodiskstat)
{
	if (Localmode) {
		FILE *in;
		char bin[PATH_MAX];
		int fid, err;

		dbug("Execname %s\n",Execname);
		strcpy(bin,dirname(Execname));
		dbug("Execdir %s\n",bin);
		strcat(bin,"/../../sbin/sysstatd");
		if (access(bin,X_OK) != 0) {
			error("cannot find sysstatd at %s",bin);
			return -1;
		}
		strcat(bin," -l");
		if (nouserstat)
			strcat(bin," -u");
		if (nodiskstat)
			strcat(bin," -D");
		if (IntervalArg)
			sprintf(bin+strlen(bin)," %s",IntervalArg);
		dbug("attaching to %s\n",bin);
		in = popen(bin,"r");
		if (in == 0)
			error("could not attach sysstatd locally: %s\n",strerror(errno));
		fid = fileno(in);
		err = fcntl(fid,F_SETFL,O_NONBLOCK);
		if (-1 == err)
			warning("could not set pipe to nonblocking: %s\n",strerror(errno));
		dbug("successfully attached to sysstatd via stream %p (fileno %d)\n",in,fid);
		return fid;
	} else {
		struct sockaddr_in MA;
		struct ip_mreq mreq;
		int on = 1, fl;
		int sock;

		sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (-1 == sock)
			error("cannot create socket: %s\n",strerror(errno));
		fl = fcntl(sock,F_GETFL,0);
		if (-1 == fcntl(sock,F_SETFL,fl | O_NONBLOCK)) 
			warning("could not set socket to nonblocking: %s\n",strerror(errno));
		if (-1 == setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)))
			error("cannot set socket to reuse address: %s\n",strerror(errno));
		bzero(&MA,sizeof(MA));
		MA.sin_family = AF_INET;
		MA.sin_port = htons(Port);
#ifdef __CYGWIN__
		MA.sin_addr.s_addr = htonl(INADDR_ANY);
		mreq.imr_multiaddr.s_addr = inet_addr(McastAddr);
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
#else
		if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
			error("cannot set multicast address: %s\n",strerror(errno));
		memcpy(&mreq.imr_multiaddr.s_addr,(void *)&MA.sin_addr,sizeof(struct in_addr));
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);
#endif
		if (-1 == bind(sock,(struct sockaddr *)&MA,sizeof(MA)))
			error("cannot bind socket: %s\n",strerror(errno));

		if (-1 == setsockopt(sock,IPPROTO_IP,IP_ADD_MEMBERSHIP,&mreq,sizeof(mreq)))
			error("could not join multicast group: %s\n",strerror(errno));
		return sock;
	}
}


static unsigned long long
get_unsigned(const char *restrict key, const char *restrict value)
{
	unsigned long long ret = 0;
	char c;
	assert(value);
	c = *value;
	if ((c < '0') || (c > '9')) {
		warning("expected an unsigned integer, got: %s\t%s\n",key,value);
		return 0;
	}
	do {
		c -= '0';
		++value;
		ret *= 10;
		ret += c;
		c = *value;
	} while ((c >= '0') && (c <= '9'));
	return ret;
}


static void
parse_dev(char *restrict d, sysstat_t *restrict s)
{
	char *dev = d;
	fsstat_t *fs = calloc(1,sizeof(fsstat_t));
	if (s->fs) {
		fsstat_t *l = s->fs;
		while (l->next) l = l->next;
		l->next = fs;
	} else {
		s->fs = fs;
	}
	d = strchr(d,'\t');
	assert(d);
	*d = 0;
	++d;
	fs->dev = strdup(dev);
	fs->ops_rate = -1;
	//dbug("parse_dev(%s,%p)\n",dev,s);
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_dev(): key = \"%s\", value = \"%s\"\n",key,value);
		if (0 == strcmp("mnt",key)) {
			fs->mnt = strdup(value);
		} else if (0 == strcmp("fs",key)) {
			if (strlen(value) > 7) 
				warning("protocol error: %s\t%s\n",key,value);
			else
				strcpy(fs->typ,value);
		} else if (0 == strcmp("fr",key)) {
			fs->free = get_unsigned(key,value); 
		} else if (0 == strcmp("bl",key)) {
			fs->blocks = get_unsigned(key,value); 
		} else if (0 == strcmp("nz",key)) {
			fs->numzfs = get_unsigned(key,value);
		} else if (0 == strcmp("ops",key)) {
			fs->ops_rate = get_double(value); 
		} else if (0 == strcmp("inr",key)) {
			fs->read_rate = get_double(value); 
		} else if (0 == strcmp("outr",key)) {
			fs->write_rate = get_double(value); 
		} else
			warning("unkown fs subfield: %s\t%s\n",key,value);
	}
	if (fs->mnt == 0)
		fs->mnt = strdup("");
}


static void
parse_disk(char *restrict d, sysstat_t *restrict s)
{
	diskrates_t *r = calloc(1,sizeof(diskrates_t));
	char *disk = d;
	d = strchr(d,'\t');
	*d = 0;
	++d;
	assert(strlen(disk) < sizeof(r->device));
	strcpy(r->device,disk);
	//dbug("parse_disk('%s',%p)\n",d,s);
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_disk(%s): key = \"%s\", value = \"%s\"\n",r->device,key,value);
		if (0 == strcmp("nread",key)) {
			r->nread = get_double(value);
		} else if (0 == strcmp("nwritten",key)) {
			r->nwritten = get_double(value);
		} else if (0 == strcmp("reads",key)) {
			r->reads = get_double(value);
		} else if (0 == strcmp("writes",key)) {
			r->writes = get_double(value);
		} else if (0 == strcmp("rbusy",key)) {
			r->rbusy = get_double(value);
		} else if (0 == strcmp("wbusy",key)) {
			r->wbusy = get_double(value);
		} 
	}
	r->next = s->diskrates;
	s->diskrates = r;
}


static void
parse_net(char *restrict d, sysstat_t *restrict s)
{
	netrates_t *r = calloc(1,sizeof(netrates_t));
	char *net = d;
	d = strchr(d,'\t');
	*d = 0;
	++d;
	assert(strlen(net) < sizeof(r->device));
	strcpy(r->device,net);
	//dbug("parse_net('%s',%p)\n",d,s);
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_net(%s): key = \"%s\", value = \"%s\"\n",r->device,key,value);
		if (0 == strcmp("ipkt",key)) {
			r->ipckts = get_double(value);
		} else if (0 == strcmp("opkt",key)) {
			r->opckts = get_double(value);
		} else if (0 == strcmp("inb",key)) {
			r->inbs = get_double(value);
		} else if (0 == strcmp("outb",key)) {
			r->outbs = get_double(value);
		} 
	}
	r->next = s->netrates;
	s->netrates = r;
}


static void
parse_user(char *restrict d, sysstat_t *restrict s)
{
	userstat_t *l = 0;
	userstat_t *us = calloc(1,sizeof(userstat_t));
	if (1 != sscanf(d,"%lu",(long unsigned *) &us->uid)) {
		warning("protocol error: uid\t%s\n",d);
		free(us);
		return;
	}
	if (s->us) {
		l = s->us;
		while (l->next) l = l->next;
		l->next = us;
	} else {
		s->us = us;
	}
	d = strchr(d,'\t');
	assert(d);
	++d;
	while (d && *d) {
		char *key = d, *value;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		//dbug("parse_user(): key = \"%s\", value = \"%s\"\n",key,value);
		if (*(key + 1) == 0) {
			switch (*key) {
			case 'N':
				us->nproc = get_unsigned(key,value); 
				break;
			case 'L':
				us->nlwp = get_unsigned(key,value); 
				break;
			case 'O':
				us->oncpu = get_unsigned(key,value); 
				break;
			case 'R':
				us->runnable = get_unsigned(key,value); 
				break;
			case 'S':
				us->sleeping = get_unsigned(key,value); 
				break;
			case 'T':
				us->stopped = get_unsigned(key,value); 
				break;
			case 'Z':
				us->zombie = get_unsigned(key,value); 
				break;
			default:
				warning("unknown single char key %c\n",*key);
			}
		} else if (0 == strcmp("user",key)) {
			us->username = strdup(value);
		} else if (0 == strcmp("sec",key)) {
			if (sscanf(value,"%ld",&us->time.tv_sec) != 1) 
				warning("sec field error: %s\t%s\n",key,value);
		} else if (0 == strcmp("RSS",key)) {
			us->rss = get_unsigned(key,value); 
		} else if (0 == strcmp("SIZE",key)) {
			us->size = get_unsigned(key,value); 
		} else if (0 == strcmp("HS",key)) {
			us->h_s = get_unsigned(key,value); 
		} else if (0 == strcmp("%C",key)) {
			us->pctcpu = (unsigned short)((get_double(value)/100)*0x8000); 
		} else if (0 == strcmp("%M",key)) {
			us->pctmem = (unsigned short)((get_double(value)/100)*0x8000); 
		} else
			warning("unkown user subfield: %s\t%s\n",key,value);
	}
	if (us->username == 0)
		us->username = strdup("<unknown>");
}


static void
parse_cpurates(char *restrict d, sysstat_t *restrict s)
{
	int id;
	cpurates_t *r;

	id = get_unsigned(0,d);
	r = s->cpurates;
	while (r && (r->id != id)) {
		r = r->next;
	}
	if (r == 0) {
		r = calloc(1,sizeof(cpurates_t));;
		r->next = s->cpurates;
		s->cpurates = r;
	}
	r->id = id;
	d = strchr(d,'\t');
	assert(d);
	*d = 0;
	++d;
	while (d && *d) {
		char *key = d, *value;
		unsigned l;
		value = strchr(d,'\t');
		assert(value);
		*value = 0;
		++value;
		d = strchr(value,'\t');
		if (d) {
			*d = 0;
			++d;
		}
		l = (value-key)-1;
		if (l == 4) {
			if (0 == strcmp("sysc",key))
				r->syscall = get_double(value);
			else if (0 == strcmp("intr",key))
				r->intr = get_double(value);
			else if (0 == strcmp("ithr",key))
				r->intrthread = get_double(value);
			else if (0 == strcmp("iblk",key))
				r->intrblk = get_double(value);
			else if (0 == strcmp("xcal",key))
				r->xcalls = get_double(value);
			else if (0 == strcmp("migr",key))
				r->migrate = get_double(value);
			else if (0 == strcmp("icsw",key))
				r->inv_swtch = get_double(value);
			else if (0 == strcmp("smtx",key))
				r->mxae = get_double(value);
			else if (0 == strcmp("rwrf",key))
				r->rwrf = get_double(value);
			else if (0 == strcmp("rwwf",key))
				r->rwwf = get_double(value);
			else if (0 == strcmp("hatf",key))
				r->hat_fault = get_double(value);
			else if (0 == strcmp("majf",key))
				r->majfault = get_double(value);
			else
				warning("unknown cpu rates subfield \"%s\"\n",key);
		} else if (l == 3) {
			if (0 == strcmp("usr",key))
				r->user = get_double(value);
			else if (0 == strcmp("sys",key))
				r->kernel = get_double(value);
			else if (0 == strcmp("idl",key))
				r->idle = get_double(value);
			else if (0 == strcmp("csw",key))
				r->pswitch = get_double(value);
			else if (0 == strcmp("asf",key))
				r->as_fault = get_double(value);
			else
				warning("unknown cpu rates subfield \"%s\"\n",key);
		} else {
			warning("unknown cpu rates subfield \"%s\"\n",key);
		}
	}
	dbug("received cpurates for cpu %d\n",r->id);
}


static char *get_pair(char *restrict str, char **restrict key, char **restrict value)
{
	if ((str == 0) || (*str == 0))
		return 0;
	*key = str;
	str = strchr(str,'\t');
	if (str == 0) {
		warning("key without value (%s)\n",*key);
		*key = 0;
		return 0;
	}
	*str = 0;
	++str;
	*value = str;
	str = strchr(*value,'\n');
	if (str == 0) {
		warning("unterminated key %s\n",*key);
		*key = 0;
		*value = 0;
		return 0;
	}
	*str = 0;
	return ++str;
}


sysstat_t *
receive_stats(int sock)
{
	static char msg[4172+1], *m;	/* maximum size of one packet plus \0 */
	/* key and value are pointers into msg */
	char *key,*value;
	int num;
	struct sockaddr_in from;
	socklen_t len;
	sysstat_t *s;

	if (Localmode)
		num = read(sock,&msg,sizeof(msg));
	else
#if defined(__CYGWIN__)
		num = recvfrom(sock,&msg,sizeof(msg)-1,0,0,0);
#else
		num = recvfrom(sock,&msg,sizeof(msg)-1,0,(struct sockaddr *)&from,&len);
#endif
	if (num == -1) {
		if (errno == EWOULDBLOCK)
			return 0;
		error("cannot read from socket: %s\n",strerror(errno));
	}
	dbug("received %d bytes\n",num);
	msg[num] = '\0';
	dbug("packet:\n%s\n",msg);
	m = get_pair(msg,&key,&value);
	if (key == 0)
		error("cannot attach to sysstatd\n");
	if (strcmp(key,"sysstat") || strcmp(value,"1")) {
		warning("invalid protcol\n");
		return 0;
	}
	s = calloc(1,sizeof(sysstat_t));
	assert(s);
	s->NetInB = s->NetOutB = s->IpInR = s->IpOutR = -1;
	while ((m = get_pair(m,&key,&value)) != 0) {
		//dbug("key: \"%s\", value: \"%s\"\n",key,value);
		switch (*key++) {
		case 'a':
			if (strcmp("anonfree"+1,key) == 0)
				s->AnonfreeR = get_double(value);
			else if (strcmp("anonpgin"+1,key) == 0)
				s->AnonpginR = get_double(value);
			else if (strcmp("anonpgout"+1,key) == 0)
				s->AnonpgoutR = get_double(value);
			else
				break;
			continue;

		case 'B':
			if (strcmp("BT"+1,key) == 0)
				s->BootTime = get_unsigned(key,value);
			else
				break;
			continue;

		case 'c':
			if (strcmp("cpu"+1,key) == 0)
				parse_cpurates(value,s);
			else if (strcmp("cpumigrate"+1,key) == 0)
				s->MigrateR = get_double(value);
			else if (strcmp("call"+1,key) == 0)
				s->SyscallR = get_double(value);
			else
				break;
			continue;

		case 'd':
			if (strcmp("dev"+1,key) == 0) {
				parse_dev(value,s);
				++s->numfs;
			} else if (strcmp("disk"+1,key) == 0) {
				parse_disk(value,s);
				++s->numdisks;
			} else
				break;
			continue;

		case 'D':
			if (strcmp("DiskI"+1,key) == 0)
				s->DiskIR = get_double(value);
			else if (strcmp("DiskO"+1,key) == 0)
				s->DiskOR = get_double(value);
			else if (strcmp("Disk"+1,key) == 0)
				s->Devices = get_unsigned(key,value);
			else
				break;
			continue;

		case 'e':
			if (strcmp("exec"+1,key) == 0)
				s->SysexecR = get_double(value);
			else
				break;
			continue;

		case 'f':
			if (strcmp("fork"+1,key) == 0)
				s->SysforkR = get_double(value);
			else
				break;
			continue;

		case 'F':
			if (strcmp("Free"+1,key) == 0)
				s->Pagesfree = get_unsigned(key,value);
			else if (strcmp("FsFree"+1,key) == 0)
				s->FsFree = get_unsigned(key,value);
			else
				break;
			continue;

		case 'H':
			if (strcmp("Host"+1,key) == 0) {
				if (strlen(value) > sizeof(s->Hostinfo))
					warning("hostname to long: %s\t%s\n"+1,key,value);
				else
					strcpy(s->Hostinfo,value);
			} else if (strcmp("HErr"+1,key) == 0)
				s->HardErrors = get_unsigned(key,value);
			else
				break;
			continue;

		case 'I':
			if (strcmp("Idle"+1,key) == 0)
				s->IdleP = get_double(value);
			else if (strcmp("IpIErr"+1,key) == 0)
				s->IpInErrs = get_unsigned(key,value);
			else if (strcmp("IpIDis"+1,key) == 0)
				s->IpInDiscards = get_unsigned(key,value);
			else if (strcmp("IpODis"+1,key) == 0)
				s->IpOutDiscards = get_unsigned(key,value);
			else if (strcmp("IpI"+1,key) == 0)
				s->IpInR = get_double(value);
			else if (strcmp("IpO"+1,key) == 0)
				s->IpOutR = get_double(value);
			else if (strcmp("Itv"+1,key) == 0)
				(void) key;	/* do nothing, make lint happy */
			else
				break;
			continue;

		case 'i':
			if (strcmp("intr"+1,key) == 0)
				s->IntrR = get_double(value);
			else if (strcmp("intrblk"+1,key) == 0)
				s->IntrBlkR = get_double(value);
			else if (strcmp("inv_swtch"+1,key) == 0)
				s->InvSwtchR = get_double(value);
			else
				break;
			continue;

		case 'K':
			if (strcmp("Kernel"+1,key) == 0)
				s->KernelP = get_double(value);
			else
				break;
			continue;

		case 'L':
			if (strcmp("Load1"+1,key) == 0) 
				s->loadavg[0] = get_double(value);
			else if (strcmp("Load5"+1,key) == 0) 
				s->loadavg[1] = get_double(value);
			else if (strcmp("Load15"+1,key) == 0) 
				s->loadavg[2] = get_double(value);
			else if (strcmp("Locked"+1,key) == 0)
				s->Pageslocked = get_unsigned(key,value);
			else
				break;
			continue;

		case 'l':
			if (strcmp("lockf"+1,key) == 0)
				s->LockfR = get_double(value);
			else
				break;
			continue;

		case 'N':
			if (strcmp("NCPU"+1,key) == 0)
				s->Ncpus = get_unsigned(key,value);
			else if (strcmp("NetI"+1,key) == 0)
				s->NetInR = get_double(value);
			else if (strcmp("NetO"+1,key) == 0)
				s->NetOutR = get_double(value);
			else if (strcmp("NetIb"+1,key) == 0)
				s->NetInB = get_double(value);
			else if (strcmp("NetOb"+1,key) == 0)
				s->NetOutB = get_double(value);
			else if (strcmp("NfsI"+1,key) == 0)
				s->NfsIR = get_double(value);
			else if (strcmp("NfsO"+1,key) == 0)
				s->NfsOR = get_double(value);
			else if (strcmp("NfsD"+1,key) == 0)
				s->NfsdR = get_double(value);
			else
				break;
			continue;
		case 'n':
			if (strcmp("net"+1,key) == 0)
				parse_net(value,s);
			else
				break;
			continue;
		
		case 'O':
			if (strcmp("OCPU"+1,key) == 0)
				s->Ocpus = get_unsigned(key,value);
			else
				break;
			continue;
			
		case 'p':
			if (strcmp("pgswapin"+1,key) == 0)
				s->PgswapinR = get_double(value);
			else if (strcmp("pgswapout"+1,key) == 0)
				s->PgswapoutR = get_double(value);
			else if (strcmp("pswitch"+1,key) == 0)
				s->PswitchR = get_double(value);
			else
				break;
			continue;
			
		case 'r':
			if (strcmp("read"+1,key) == 0)
				s->SysreadR = get_double(value);
			else
				break;
			continue;
			
		case 'R':
			if (strcmp("Ram"+1,key) == 0)
				s->Physmem = get_unsigned(key,value);
			else
				break;
			continue;

		case 's':
			if (strcmp("scan"+1,key) == 0)
				s->ScanR = get_double(value);
			else
				break;
			continue;
			
		case 'S':
			if (strcmp("SwapTotal"+1,key) == 0)
				s->SwapTotal = get_unsigned(key,value);
			else if (strcmp("SwapFree"+1,key) == 0)
				s->SwapFree = get_unsigned(key,value);
			else if (strcmp("SErr"+1,key) == 0)
				s->SoftErrors = get_unsigned(key,value);
			else if (strcmp("Storage"+1,key) == 0)
				s->Storage = get_unsigned(key,value);
			else
				break;
			continue;

		case 'T':
			if (strcmp("TS"+1,key) == 0)
				s->TS = get_unsigned(key,value);
			else
				break;
			continue;

		case 'U':
			if (strcmp("User"+1,key) == 0)
				s->UserP = get_double(value);
			else
				break;
			continue;

		case 'u':
			if (strcmp("uid"+1,key) == 0)
				parse_user(value,s);
			else
				break;
			continue;

		case 'w':
			if (strcmp("write"+1,key) == 0)
				s->SyswriteR = get_double(value);
			else
				break;
			continue;

		case 'x':
			if (strcmp("xcall"+1,key) == 0)
				s->XcallsR = get_double(value);
			else
				break;
			continue;

		case 'Z':
			if (strcmp("Zpools"+1,key) == 0)
				s->Zpools = get_unsigned(key,value);
			else if (strcmp("Zsize"+1,key) == 0)
				s->Zsize = get_unsigned(key,value);
			else if (strcmp("Zfree"+1,key) == 0)
				s->Zfree = get_unsigned(key,value);
			else if (strcmp("ZfsArcSize"+1,key) == 0)
				s->ZfsArcSize = get_unsigned(key,value);
			else
				break;
			continue;

		default:
			;
		}
		info("ignoring unknown key: %s\n",key);
	}
	return s;
}

// vim:tw=0
