/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include "fsstats.h"
#include "log.h"

#include <alloca.h>
#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <link.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/statvfs.h>
#include <unistd.h>

#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

extern int VerboseZFS;

static fsstat_t *Fs = 0;

typedef struct libzfs_handle libzfs_handle_t;
typedef struct zpool_handle zpool_handle_t;
typedef int (*zpool_iter_f)(zpool_handle_t *, void *);

static void *Libzfs = 0;
static libzfs_handle_t *libzfs_handle = 0;
static libzfs_handle_t *(*libzfs_init)(void) = 0;
static int (*zpool_iter)(zpool_iter_f,void *) = 0;
static int (*zpool_iter2)(libzfs_handle_t *, zpool_iter_f, void *) = 0;
static uint64_t (*zpool_get_space_total)(zpool_handle_t *) = 0;
static uint64_t (*zpool_get_space_used)(zpool_handle_t *) = 0;
static uint64_t (*zpool_get_prop)(zpool_handle_t *,int,char *,int,int *) = 0;
static uint64_t (*zpool_get_prop_int)(zpool_handle_t *,int,int *) = 0;
static char *(*zpool_get_name)(zpool_handle_t *) = 0;
static int (*zpool_get_root)(zpool_handle_t *, char *, size_t) = 0;
static void (*zpool_close)(zpool_handle_t *) = 0;



void init_zfs(void)
{
	Libzfs = dlopen("libzfs.so",RTLD_NOW|RTLD_GLOBAL);
	if (0 == Libzfs)
		return;
	libzfs_init = (libzfs_handle_t *(*)())dlsym(Libzfs,"libzfs_init");
	if (libzfs_init) {
		libzfs_handle = libzfs_init();
		zpool_iter2 = (int (*)())dlsym(Libzfs,"zpool_iter");
		assert(zpool_iter2);
	} else {
		zpool_iter = (int (*)())dlsym(Libzfs,"zpool_iter");
		assert(zpool_iter);
	}
	zpool_close = (void (*)()) dlsym(Libzfs,"zpool_close");
	assert(zpool_close);
	zpool_get_space_total = (uint64_t (*)()) dlsym(Libzfs,"zpool_get_space_total");
	if (zpool_get_space_total) {
		zpool_get_space_used = (uint64_t (*)()) dlsym(Libzfs,"zpool_get_space_used");
		assert(zpool_get_space_used);
		zpool_get_root = (int (*)()) dlsym(Libzfs,"zpool_get_root");
		assert(zpool_get_root);
	} else {
		zpool_get_prop = (uint64_t (*)()) dlsym(Libzfs,"zpool_get_prop");
		assert(zpool_get_prop);
		zpool_get_prop_int = (uint64_t (*)()) dlsym(Libzfs,"zpool_get_prop_int");
		assert(zpool_get_prop_int);
	}
	zpool_get_name = (char * (*)()) dlsym(Libzfs,"zpool_get_name");
	assert(zpool_get_name);
}


static fsstat_t *
add_fs(const char *restrict dev, const char *restrict mnt, const char *restrict typ,
	uint64_t blocks, unsigned long long free, unsigned long id)
{
	fsstat_t *f = Fs;

	while (f && (strcmp(dev,f->dev) || strcmp(typ,f->typ)))
		f = f->next;
	if (f == 0) {
		dbug("add_fs(%s,%s,%s,%llu,%llu,%lx)\n",dev,mnt,typ,blocks,free,id);
		f = malloc(sizeof(fsstat_t));
		bzero(f,sizeof(fsstat_t));
		f->next = Fs;
		f->dev = strdup(dev);
		f->mnt = strdup(mnt);
		f->fsid = id;
		(void) strcpy(f->typ,typ);
		f->ops_rate = -1;
		Fs = f;
	} else
		dbug("update_fs(%s,%s,%s,%llu,%llu,%lx)\n",dev,mnt,typ,blocks,free,id);
	f->blocks = blocks;
	f->free = free;
	return f;
}


#ifdef DEBUG
void
print_fsstat(void)
{
	fsstat_t *f = Fs;
	while (f) {
		printf("%s@%s (%s): %llu\n",f->dev,f->mnt,f->typ,(long long unsigned) f->blocks);
		f = f->next;
	}
}
#endif


static int
analyze_zpool(zpool_handle_t *zh, void *ignored)
{
	uint64_t s, f;
	char root[PATH_MAX] = "", *name;
	struct statvfs st;
	int err;
	
	assert(ignored == 0);
	name = zpool_get_name(zh);
	if (zpool_get_prop) {
		int src = 0x2 /* ZFS_PROP_SRC_DEFAULT */;
		s = zpool_get_prop_int(zh,1 /* ZPOOL_PROP_SIZE */,&src) >> 10;
		f = zpool_get_prop_int(zh,3 /* ZPOOL_PROP_AVAILABLE */,&src) >> 10;
		err = zpool_get_prop(zh,5 /* ZFS_PROP_ALTROOT */,root,PATH_MAX,0);
		assert(err == 0);
	} else if (zpool_get_space_total) {
		s = zpool_get_space_total(zh) >> 10;
		f = s - (zpool_get_space_used(zh) >> 10);
		zpool_get_root(zh,root,PATH_MAX);
	} else {
		abort();
	}
	if ((root[0] == 0) || ((root[0] == '-') && (root[1] == 0))) {
		root[0] = '/';
		(void) strcpy(root+1,name);
	}
	if (statvfs(root,&st) != -1) {
		size_t l = strlen(name);
		unsigned num = 0;
		fsstat_t *fs, *z;
		z = add_fs(name,root,"zpool",s,f,st.f_fsid);
		z->ops_rate = 0;
		z->read_rate = 0;
		z->write_rate = 0;
		fs = Fs;
		while (fs) {
			if ((strncmp(name,fs->dev,l) == 0) && (fs->dev[l] == '/')) {
				fs->zroot = z;
				++num;
				dbug("associated filesystem %s with pool %s\n",fs->dev,name);
			}
			fs = fs->next;
		}
		dbug("pool %s has %d filesystems\n",name,num);
		z->numzfs = num;
	}
	zpool_close(zh);
	return 0;
}


static void
get_zfs_stats(void)
{
	int err;
	if (libzfs_init) {
		err = zpool_iter2(libzfs_handle,analyze_zpool,0);
	} else {
		err = zpool_iter(analyze_zpool,0);
	}
	assert(err == 0);
}


static int
Crit(fsstat_t *restrict a, fsstat_t *restrict b)
{
	if (b == 0)
		return 1;
	assert(a);
	if (a->free == b->free)
		return strcmp(a->dev,b->dev) < 0;
	return a->free > b->free;
}


static void
sort_fs(void)
{
	int del = 0;
	fsstat_t *f = Fs, *h = 0, *t = 0;

	f = Fs;
	while (f) {
		fsstat_t *n = f->next, *i = h;
		if (f->blocks == 0) {
			++del;
			free(f->dev);
			free(f->mnt);
			free(f);
			f = n;
			continue;
		}
		while (i && Crit(f,i)
			&& i->next && Crit(f,i->next)) {
			assert(i != i->next);
			i = i->next;
		}
		if (i == 0) {
			if (t) {
				assert(h && !Crit(t, i));
				t->next = f;
				t = f;
			} else {
				assert(h == 0);
				h = t = f;
			}
			f->next = 0;
		} else {
			if (Crit(f,i)) {
				f->next = i->next;
				i->next = f;
			} else {
				f->next = i;
				h = f;
			}
			assert(i != i->next);
		}
		f = n;
		assert(f == 0 || f != f->next);
	}
	Fs = h;
#ifdef DEBUG
	{
		int n = 0;
		f = Fs;
		dbug("sort_fs(): sorted file systems\n");
		while (f) {
			dbug("%s:%lld\n",f->dev,f->free>>20);
			++n;
			f = f->next;
		}
		dbug("sort_fs(): %d filesystems remaining, %d deleted\n",n,del);
	}
#endif
}


int
update_fs_stats(void)
{
	static int numnfs = 0;
	static time_t oldstat = 0;
	struct statvfs vst;
	char device[PATH_MAX], mnt_pnt[PATH_MAX],
	     fstyp[16];	/* Is sizeof fstyp OK? */
	int nfscount = 0, err;
	struct stat st;

	err = stat("/etc/mnttab",&st);
	assert(err == 0);
	if (oldstat != st.st_mtime) {
		fsstat_t *f = Fs;
		char *buf;
		int mntfd, num;
		oldstat = st.st_mtime;
		buf = alloca(st.st_size+1);
		dbug("rereading /etc/mnttab\n");
		mntfd = open("/etc/mnttab",O_RDONLY);
		if (-1 == mntfd) {
			error("unable to open /etc/mnttab for reading: %s\n",strerror(errno));
		}
		num = pread(mntfd,buf,st.st_size,0);
		(void) close(mntfd);
		assert(num >= 0);
		buf[num] = 0;
		while (f) {
			f->blocks = 0;	/* marks fs that have been umounted for
					   removal */
			f = f->next;
		}
		while (buf) {
			uint64_t size, avail;
			err = sscanf(buf,"%[^\t]\t%[^\t]\t%[^\t]\t%*[^\n]\n", device, mnt_pnt, fstyp);
			if (err < 3)
				break;
			buf = strchr(buf,'\n');
			if (buf)
				++buf;
			if (0 == memcmp("nfs",fstyp,4)) {
				++nfscount;
				continue;
			}
			if ((0 != memcmp("ufs",fstyp,4)) && (0 != memcmp("pcfs",fstyp,5))
					&& (0 != memcmp("tmpfs",fstyp,6)) && (0 != memcmp("vxfs",fstyp,5))
					&& (0 != memcmp("zfs",fstyp,4)))
				continue;
			if (-1 == statvfs(mnt_pnt,&vst))
				continue;
			size = ((uint64_t) vst.f_blocks * vst.f_frsize) >> 10;
			dbug("%s @ %s : %llu/%llu\n",
					fstyp,
					mnt_pnt,
					(uint64_t)vst.f_bfree,
					(uint64_t)vst.f_blocks);
			avail = ((uint64_t)vst.f_bavail * vst.f_frsize) >> 10;
			(void) add_fs(device,mnt_pnt,fstyp,size,avail,vst.f_fsid);
		}
	} else {
		fsstat_t *f = Fs;
		while (f) {
			fsstat_t *n = f->next;
			if (0 == memcmp(f->typ,"zpool",6)) {
				f = n;
				continue;
			}
			if ((0 == VerboseZFS) && (0 == memcmp(f->typ,"zfs",4))) {
				f = n;
				continue;
			}
			if (-1 != statvfs(f->mnt,&vst)) {
				dbug("updated %s@%s: %llu/%llu\n",
					f->typ,
					f->mnt,
					(uint64_t)vst.f_bfree,
					(uint64_t)vst.f_blocks);
				f->blocks = ((uint64_t) vst.f_blocks * vst.f_frsize) >> 10;
				f->free = ((uint64_t)vst.f_bavail * vst.f_frsize) >> 10;
			}
			f = n;
		}
	}
	if (Libzfs)
		get_zfs_stats();
	sort_fs();
	if (nfscount != numnfs) {
		numnfs = nfscount;
		return 1;
	}
	return 0;
}


fsstat_t *
get_fs_stats(void)
{
	return Fs;
}

// vim:tw=0
