/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include <alloca.h>
#include <assert.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <math.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <kstat.h>
#include <poll.h>
#include <pthread.h>
#include <strings.h>
#include <sys/ioctl.h>
#include <sys/loadavg.h>
#include <sys/mkdev.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <sys/swap.h>
#include <sys/systeminfo.h>
#include <sys/sysinfo.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include <utmpx.h>


/* SUSv3 does not have PATH_MAX */
#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

#include "log.h"
#include "network.h"
#include "sysstat.h"

extern int Localmode, Diskstats, Userstats;
static kstat_ctl_t *KS = 0;
sysstat_t Sysstat;


typedef struct cpudata_s {
	long long 
		user, kernel, idle, wait,
		intr, intrthread, intrblk, xcalls,
		syscall, sysexec, sysfork, sysvfork,
		sysread, syswrite,
		pgswapin, pgswapout, scan,
		anonpgin, anonpgout, anonfree,
		as_fault, hat_fault, majfault,
		phread, phwrite,
		rwrf, rwwf, mxae,
		inv_swtch, pswitch, migrate;
	hrtime_t
		user_ts, kernel_ts, idle_ts, wait_ts,
		timestamp;
	kstat_t 
		*cpustat, *cpusys;
	cpurates_t
		*rates;
	unsigned
		sys_iter, stat_iter;
} cpudata_t;


typedef struct diskdata_s {
	struct diskdata_s *next;
	kstat_t *kstat;
	unsigned iteration;
	hrtime_t timestamp;
	kstat_io_t iodata;
	diskrates_t rates;
} diskdata_t;


typedef struct netdata_s {
	struct netdata_s *next;
	unsigned long long value[4];
	kstat_t *kstat;
	unsigned iteration;
	hrtime_t timestamp;
	netrates_t rates;
} netdata_t;


typedef struct class_chain_s {
	struct class_chain_s *next;
	kstat_t *kstat;
	unsigned long long value[4];
	hrtime_t timestamp;
	unsigned iteration;
} class_chain_t;


typedef struct diskinfo {
	struct diskinfo *next;
	char dev[PATH_MAX];
	int maj,min;
} diskinfo_t;

static diskinfo_t *volatile DiskInfo = 0;

/* the number of cpus can change dynamically under Solaris */
static long
	Maxid = 0,	/* highest id of a cpu */
	Iteration = 1;	/* used to make sure data is not update twice after a reconfig */

static cpudata_t
	**Cpus = 0;
static cpurates_t
	*Cpurates = 0;
static diskrates_t
	*Diskrates = 0;
static netrates_t
	*Netrates = 0;
static kstat_t
	*Syspages = 0, *Ipstat = 0, *ZfsArc = 0;

static class_chain_t
	*Err = 0, *Nfs = 0, *NfsServer = 0;

static diskdata_t
	*Disks = 0;
static netdata_t
	*Nets = 0;

static unsigned long Pgsz;

/* ip */
static unsigned long long IpInReceives, IpOutRequests;
static hrtime_t IpIn_ts, IpOut_ts;


static int
major_id(const char *name)
{
	static char *buf = 0;
	char *off;
	int major, num;

	if (buf == 0) {
		struct stat st;
		int n2m, err;
		n2m = open("/etc/name_to_major",O_RDONLY);
		if (n2m == -1)
			error("unable to open /etc/name_to_majore: %s\n",strerror(errno));
		err = fstat(n2m,&st);
		assert(err == 0);
		buf = malloc(st.st_size);
		err = read(n2m,buf,st.st_size);
		assert(err == st.st_size);
		(void) close(n2m);
	}
	off = buf;
	do {
		off = strstr(off,name);
		assert(off);
		off += strlen(name);
	} while ((*off != ' ') && (*off != '\t'));
	num = sscanf(off,"%d\n",&major);
	assert(num == 1);
	dbug("major_id(%s) = %d\n",name,major);
	return major;
}


static void *
get_disk_infos(void *ignored)
{
	static DIR *dskdir = 0;
	char dev[PATH_MAX] = "/dev/dsk/";
	struct dirent *d;
	diskinfo_t *disks = DiskInfo;

	(void) ignored;
	dskdir = opendir(dev);
	for (;;) {
		while ((d = readdir(dskdir)) != 0) {
			struct stat st;
			diskinfo_t *disk = disks;
			hrtime_t start;

			if (memcmp(d->d_name+strlen(d->d_name)-2,"s0",3))
				continue;
			while (disk && strcmp(disk->dev,d->d_name))
				disk = disk->next;
			if (disk == 0) {
				disk = malloc(sizeof(diskinfo_t));
				disk->next = disks;
				disks = disk;
				strcpy(disk->dev,d->d_name);
			} else if (disk->min < 0) {
				--disk->min;
				if (disk->min > -100)
					continue;
			}
			strcpy(dev+9,d->d_name);
			start = gethrtime();
			if (-1 == stat(dev,&st)) {
				hrtime_t end = gethrtime();
				dbug("stat(%s,...) = -1\n",dev);
				disk->min = 0;
				disk->maj = -1;
				if (errno != ENOENT)
					warning("could not stat %s: %s\n",dev,strerror(errno));
				else if (end - start > 1000000000ULL) 
					disk->min = -1;
			} else {
				disk->min = minor(st.st_rdev);
				disk->maj = major(st.st_rdev);
				dbug("stat(%s,...): %d/%d\n",dev,disk->maj,disk->min);
				assert((st.st_mode & S_IFBLK) == S_IFBLK);
			}
		}
		DiskInfo = disks;
		rewinddir(dskdir);
		(void) sleep(60);
	}
}


static void
gather_disk_infos(void)
{
	static int num = 1;
	
	if (0 == --num) {
		int ret;
		pthread_t disknames;
		pthread_attr_t attr;

		if (Diskstats == 0)
			return;
		ret = pthread_attr_init(&attr);
		assert(ret == 0);
		ret = pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
		assert(ret == 0);
		ret = pthread_create(&disknames,&attr,get_disk_infos,0);
		assert(ret == 0);
	} else 
		abort();
}


static int
get_disk_name(const char *restrict driver, int major, int id, char *restrict name)
{
	diskinfo_t *disk = DiskInfo;
	
	if (major == -1)
		major = major_id(driver);
#ifdef __sparc
	id <<= 3;
#elif defined __i386 || defined __amd64
	id <<= 6;
#else
#error unknown architecture
#endif
	while (disk) { 
		dbug("get_disk_name(%s,%d,...): %d/%d\n",driver,id,disk->maj,disk->min);
		if (disk->maj == major && disk->min == id) {
			if (name[0] == 0) {
				size_t l = strlen(disk->dev);
				memcpy(name,disk->dev,l < 32 ? l+1 : 31);
				name[31] = 0;
				if (name[l-2] == 's')
					name[l-2] = 0;
			}
			dbug("get_disk_name(%s,%u): %s\n",driver,id,name);
			return major;
		}
		disk = disk->next;
	}
	dbug("stale disk %s%d\n",driver,id);
	return -1;
}


static void
get_boottime(void)
{
	struct utmpx utmpx, *ent;
	utmpx.ut_type = BOOT_TIME;
	ent = getutxid(&utmpx);
	assert(ent);
	Sysstat.BootTime = ent->ut_tv.tv_sec;
}


static void
get_swapinfo(void)
{
	swaptbl_t *swaptbl;
	char *strtbl;
	int numswap,i;

	numswap = swapctl(SC_GETNSWP,0);
	assert(numswap != -1);
	numswap += 2;
	swaptbl = alloca(numswap * sizeof(swapent_t) + sizeof(struct swaptable));
	swaptbl->swt_n = numswap;
	strtbl = alloca(numswap * PATH_MAX);
	for (i = 0; i < numswap; ++i)
		swaptbl->swt_ent[i].ste_path = strtbl + (i * PATH_MAX);
	numswap = swapctl(SC_LIST,swaptbl);
	/* If you encounter the following assertion, then please
	 * recompile in 64 bit mode. It is caused by an interface
	 * limitation.
	 */
	assert(numswap != -1);
	Sysstat.SwapTotal = Sysstat.SwapFree = 0;
	for (i = 0; i < numswap; ++i) {
		Sysstat.SwapTotal += swaptbl->swt_ent[i].ste_pages;
		Sysstat.SwapFree += swaptbl->swt_ent[i].ste_free;
	}
	Sysstat.SwapTotal = ((long long)Sysstat.SwapTotal * Pgsz) >> 10;
	Sysstat.SwapFree = ((long long)Sysstat.SwapFree * Pgsz) >> 10;
}


static class_chain_t *
get_class_chain(const char *restrict class, const char *restrict module, const char *restrict name, class_chain_t *c)
{
	kstat_t *k;
	class_chain_t *p = 0, *f = c;
	size_t nl = name ? strlen(name)+1 : 0;
	size_t cl = class ? strlen(class)+1 : 0;
	size_t ml = module ? strlen(module) : 0;
	if (0 == c) {
		c = calloc(1,sizeof(class_chain_t));
		assert(c);
		f = c;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if ((class && memcmp(class,k->ks_class,cl)) ||
		    (module && memcmp(module,k->ks_module,ml)) ||
		    (name && memcmp(name,k->ks_name,nl)))
			continue;
		if (0 == c) {
			c = malloc(sizeof(class_chain_t));
			/* p is always a valid pointer, because
			 * we allocate c before going into this loop
			 * and in every iteration if necessary
			 */
			c->next = 0;
			c->iteration = 0;
			p->next = c;
		}
		c->value[0] = c->value[1] = 0;
		dbug("get_class_chain(): adding %s/%s/%s\n",k->ks_class,k->ks_module,k->ks_name);
		c->kstat = k;
		p = c;
		c = c->next;
	}
	while (c) {
		c->kstat = 0;
		c = c->next;
	}
	return f;
}


static class_chain_t *
get_chain_by_name(const char *name, class_chain_t *c)
{
	kstat_t *k;
	class_chain_t *p = 0, *f = c;
	size_t nl = strlen(name);	// ignore terminating null byte!
	if (0 == c) {
		c = calloc(1,sizeof(class_chain_t));
		assert(c);
		f = c;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if (memcmp(name,k->ks_name,nl))
			continue;
		if (0 == c) {
			c = malloc(sizeof(class_chain_t));
			/* p is always a valid pointer, because
			 * we allocate c before going into this loop
			 * and in every iteration if necessary
			 */
			c->next = 0;
			c->iteration = 0;
			p->next = c;
		}
		c->value[0] = c->value[1] = 0;
		dbug("get_class_chain(): adding %s/%s/%s\n",k->ks_class,k->ks_module,k->ks_name);
		c->kstat = k;
		p = c;
		c = c->next;
	}
	while (c) {
		c->kstat = 0;
		c = c->next;
	}
	return f;
}


static class_chain_t *
get_chain_by_class(const char *class, class_chain_t *c)
{
	kstat_t *k;
	class_chain_t *p = 0, *f = c;
	size_t cl = strlen(class) + 1;

	if (0 == c) {
		c = calloc(1,sizeof(class_chain_t));
		assert(c);
		f = c;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if (memcmp(class,k->ks_class,cl))
			continue;
		if (0 == c) {
			c = malloc(sizeof(class_chain_t));
			/* p is always a valid pointer, because
			 * we allocate c before going into this loop
			 * and in every iteration if necessary
			 */
			c->next = 0;
			c->iteration = 0;
			p->next = c;
		}
		c->value[0] = c->value[1] = 0;
		dbug("get_class_chain(): adding %s/%s/%s\n",k->ks_class,k->ks_module,k->ks_name);
		c->kstat = k;
		p = c;
		c = c->next;
	}
	while (c) {
		c->kstat = 0;
		c = c->next;
	}
	return f;
}


/* Solaris >=10 only */
static void
get_ip_ip(void)
{
	kstat_named_t *data;
	kid_t kid;

	kid = kstat_read(KS,Ipstat,0);
	assert(kid != -1);

	data = kstat_data_lookup(Ipstat,"inReceives");
	assert(data);
	if (data->data_type == KSTAT_DATA_UINT32) {
		Sysstat.IpInR = (double)(data->value.ui32 - IpInReceives) * 1E9 /
			(double)(Ipstat->ks_snaptime - IpIn_ts);
		IpIn_ts = Ipstat->ks_snaptime;
		IpInReceives = data->value.ui32;
	} else if (data->data_type == KSTAT_DATA_UINT64) {
		Sysstat.IpInR = (double)(data->value.ui64 - IpInReceives) * 1E9 /
			(double)(Ipstat->ks_snaptime - IpIn_ts);
		IpIn_ts = Ipstat->ks_snaptime;
		IpInReceives = data->value.ui64;
	} else
		assert(0);

	data = kstat_data_lookup(Ipstat,"outRequests");
	assert(data);

	if (data->data_type == KSTAT_DATA_UINT32) {
		Sysstat.IpOutR = (double)(data->value.ui32 - IpOutRequests) * 1E9/
			(double)(Ipstat->ks_snaptime - IpOut_ts);
		IpOut_ts = Ipstat->ks_snaptime;
		IpOutRequests = data->value.ui32;
	} else if (data->data_type == KSTAT_DATA_UINT64) {
		Sysstat.IpOutR = (double)(data->value.ui64 - IpOutRequests) * 1E9/
			(double)(Ipstat->ks_snaptime - IpOut_ts);
		IpOut_ts = Ipstat->ks_snaptime;
		IpOutRequests = data->value.ui64;
	} else 
		assert(0);

	data = kstat_data_lookup(Ipstat,"inDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	Sysstat.IpInDiscards = data->value.ui32;

	data = kstat_data_lookup(Ipstat,"outDiscards");
	assert(data);
	assert(data->data_type == KSTAT_DATA_UINT32);
	Sysstat.IpOutDiscards = data->value.ui32;
}


static int
get_device_errors(void)
{
	class_chain_t *c;

	Sysstat.HardErrors = 0;
	Sysstat.SoftErrors = 0;
	Sysstat.Storage = 0;
	Sysstat.Devices = 0;
	for (c = Err; c; c = c->next) {
		kstat_named_t *data;

		if (-1 == kstat_read(KS,c->kstat,0))
			return 1;
		assert(c->kstat->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(c->kstat,"Size");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		if (data->value.ui64 == 0)
			continue;
		++Sysstat.Devices;
		Sysstat.Storage += data->value.ui64;
		data = kstat_data_lookup(c->kstat,"Hard Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		Sysstat.HardErrors += data->value.ui32;
		data = kstat_data_lookup(c->kstat,"Soft Errors");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT32);
		Sysstat.SoftErrors += data->value.ui32;
	}
	Sysstat.Storage >>= 10;
	return 0;
}


static int
get_disk_io(void)
{
	diskdata_t *c;
	for (c = Disks; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		c->rates.major = get_disk_name(s->ks_module,c->rates.major,s->ks_instance,c->rates.disk);
		if (c->rates.major == -1)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert((s->ks_type == KSTAT_TYPE_IO)
			&& (s->ks_snaptime != c->timestamp));
		assert(strcmp(s->ks_name,c->rates.device) == 0);
		data = s->ks_data;
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		assert(td >= 0);
		if (c->iodata.nread)
			Sysstat.DiskIR += (double)(data->nread - c->iodata.nread) * td;
		if (c->iodata.nwritten)
			Sysstat.DiskOR += (double)(data->nwritten - c->iodata.nwritten) * td;
		c->rates.reads = (data->reads - c->iodata.reads) * td;
		c->rates.writes = (data->writes - c->iodata.writes) * td;
		c->rates.nread = (double) ((data->nread - c->iodata.nread) >> 10) * td;
		c->rates.nwritten = (double)((data->nwritten - c->iodata.nwritten) >> 10) * td;
		c->rates.rbusy = (data->rtime - c->iodata.rtime) / (double)(s->ks_snaptime - c->timestamp) * 100.0;
		dbug("rbusy: (%lld - %lld) == %lld / (%lld - %lld ) == %lld = %g\n",data->rtime,c->iodata.rtime,data->rtime-c->iodata.rtime,s->ks_snaptime,c->timestamp,s->ks_snaptime-c->timestamp,c->rates.rbusy);
		if (c->rates.rbusy < 0)
			c->rates.wbusy = 0;
		c->rates.wbusy = (data->wtime - c->iodata.wtime) / (double)(s->ks_snaptime - c->timestamp) * 100.0;
		dbug("disk %s/%s: %g %g/%g\n",c->rates.disk,c->rates.device,c->rates.reads+c->rates.writes,c->rates.nread,c->rates.nwritten);
		if (c->rates.wbusy < 0)
			c->rates.wbusy = 0;
//		c->rates.rcnt = (data->rcnt - c->iodata.rcnt) * td;
//		c->rates.wcnt = (data->wcnt - c->iodata.wcnt) * td;
//		c->rates.rlentime = data->rlentime - c->iodata.rlentime;
//		c->rates.wlentime = data->wlentime - c->iodata.wlentime;
		memcpy(&c->iodata,data,sizeof(kstat_io_t));
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	Sysstat.DiskIR /= 1024;
	Sysstat.DiskOR /= 1024;
	return 0;
}


static int
get_nfs_io(void)
{
	class_chain_t *c;
	for (c = Nfs; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_io_t *data;
		double td;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_IO);
		data = s->ks_data;
		assert(s->ks_snaptime != c->timestamp);
		td = 1E9 / (s->ks_snaptime - c->timestamp);
		if (c->value[0])
			Sysstat.NfsIR += (double)(data->nread - c->value[0]) * td;
		if (c->value[1])
			Sysstat.NfsOR += (double)(data->nwritten - c->value[1]) * td;
		c->value[0] = data->nread;
		c->value[1] = data->nwritten;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	Sysstat.NfsIR /= 1024;
	Sysstat.NfsOR /= 1024;
	for (c = NfsServer; c; c = c->next) {
		kstat_t *s = c->kstat;
		kstat_named_t *data;

		if (s == 0) {	
			/* chain element not in use */
			continue;
		}
		if (c->iteration == Iteration)
			continue;
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert(s->ks_type == KSTAT_TYPE_NAMED);
		data = kstat_data_lookup(s,"calls");
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (c->value[0])
			Sysstat.NfsdR += (double)(data->value.ui64 - c->value[0]) * 1E9 /
				(s->ks_snaptime - c->timestamp);
		c->value[0] = data->value.ui64;
		c->timestamp = s->ks_snaptime;
		c->iteration = Iteration;
	}
	return 0;
}


static int
get_unix_systempages(void)
{
	kstat_named_t *data;

	if (kstat_read(KS,Syspages,0) == -1) {
		/* cannot be unconfigured/cannot fail */
		assert(0);
	}

	data = kstat_data_lookup(Syspages,"physmem");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Sysstat.Physmem = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Sysstat.Physmem = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagesfree");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Sysstat.Pagesfree = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Sysstat.Pagesfree = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pageslocked");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Sysstat.Pageslocked = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Sysstat.Pageslocked = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);

	data = kstat_data_lookup(Syspages,"pagestotal");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Sysstat.Pagestotal = (data->value.ui64 * Pgsz) >> 10;
	else if (data->data_type == KSTAT_DATA_UINT32)
		Sysstat.Pagestotal = ((unsigned long long) data->value.ui32 * Pgsz) >> 10;
	else
		assert(0);
	return 0;
}


static int
get_zfs_arc(void)
{
	kstat_named_t *data;

	if (ZfsArc == 0)
		return 0;
	if (kstat_read(KS,ZfsArc,0) == -1) {
		/* cannot be unconfigured/cannot fail */
		assert(0);
	}
	data = kstat_data_lookup(ZfsArc,"size");
	if (data == 0)
		return 1;
	if (data->data_type == KSTAT_DATA_UINT64)
		Sysstat.ZfsArcSize = data->value.ui64 >> 10;
	else
		assert(0);
	return 0;
}


/* this is only for Solaris 10 and later */
static int
get_cpu_sys(void)
{
	int inst = 0, num = 0;
	cpurates_t *nextrates = Cpurates;

	/* not here, because for Solaris 9 this must be done in get_cpu_stat
	 * IdleP = UserP = KernelP = 0;
	 */
	for (inst = 0; inst < Maxid; ++inst) {
		kstat_named_t *data;
		cpudata_t *cpu;
		cpurates_t *rates;
		kstat_t *cpusys;
		unsigned long long user, kernel, idle;
		hrtime_t user_ts, kernel_ts, idle_ts;

		cpu = Cpus[inst];
		rates = cpu->rates;
		cpusys = cpu->cpusys;
		if (cpusys == 0)
			continue;
		if (kstat_read(KS,cpusys,0) == -1) {
			/* processor has been taken offline */
			if (rates) {
				free(rates);
				cpu->rates = 0;
			}
			Cpus[inst]->cpusys = 0;
			return 1;
		}
		++num;
		if (cpu->sys_iter == Iteration) {
			/* fields already up-to-date */
			continue;
		}
		cpu->sys_iter = Iteration;
		if (rates == 0)
			return 1;
		rates->next = nextrates;
		nextrates = rates;

		/* Every lookup here may fail, but the whole set of metrics
		 * should be committed as a transaction, because otherwise
		 * we won't be able to handle dynamic reconfiguration
		 * correctly. So first do all lookups and then write the
		 * results back. Every lookup modifies the ks_snaptime, so
		 * we need to have seperate timestamps.
		 */
		data = kstat_data_lookup(cpusys,"cpu_nsec_user");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		user = data->value.ui64;
		user_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_kernel");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		kernel = data->value.ui64;
		kernel_ts = cpusys->ks_snaptime;
		data = kstat_data_lookup(cpusys,"cpu_nsec_idle");
		if (data == 0)
			return 1;
		assert(data->data_type == KSTAT_DATA_UINT64);
		idle = data->value.ui64;
		idle_ts = cpusys->ks_snaptime;

		rates->user = (double)(user - cpu->user)
			/ (double)(user_ts - cpu->user_ts)
			* 100;
		Sysstat.UserP += rates->user / Sysstat.Ocpus;
		cpu->user = user;
		cpu->user_ts = user_ts;
		rates->kernel = (double)(kernel - cpu->kernel)
			/ (double)(kernel_ts - cpu->kernel_ts)
			* 100;
		Sysstat.KernelP += rates->kernel / Sysstat.Ocpus;
		cpu->kernel = kernel;
		cpu->kernel_ts = kernel_ts;
		rates->idle = (double)(idle - cpu->idle)
			/ (double)(idle_ts - cpu->idle_ts)
			* 100;
		Sysstat.IdleP += rates->idle / Sysstat.Ocpus;
		cpu->idle = idle;
		cpu->idle_ts = idle_ts;
	}
	Cpurates = nextrates;
	if (num >= Sysstat.Ocpus)
		return 0;
	return 1;
}


static int
get_cpu_stat(void)
{
	unsigned inst = 0, num = 0;

	for (inst = 0; inst < Maxid; ++inst) {
		cpudata_t *cpu = Cpus[inst];
		kstat_t *cpustat = cpu->cpustat;
		cpurates_t *rates = cpu->rates;
		cpu_stat_t *stat;
		double td;

		if (cpustat == 0)
			continue;
		if (-1 == kstat_read(KS,cpustat,0)) {
			/* processor was taken offline */
			return 1;
		}

		++num;
		if (cpu->stat_iter == Iteration) {
			/* this info was updated before reconfiguring */
			continue;
		}
		stat = cpustat->ks_data;
		cpu->stat_iter = Iteration;
		td = 1 / ((cpustat->ks_snaptime - cpu->timestamp) * 1E-9);

		if (rates == 0) {
			rates = calloc(1,sizeof(cpurates_t));
			cpu->rates = rates;
		}
		rates->intr = (double)(stat->cpu_sysinfo.intr - cpu->intr) * td;
		Sysstat.IntrR += rates->intr;
		cpu->intr = stat->cpu_sysinfo.intr;
		rates->intrthread = (double)(stat->cpu_sysinfo.intrthread - cpu->intrthread) * td;
		Sysstat.IntrThreadR += rates->intrthread;
		cpu->intrthread = stat->cpu_sysinfo.intrthread;
		rates->intrblk = (double)(stat->cpu_sysinfo.intrblk - cpu->intrblk) * td;
		Sysstat.IntrBlkR += rates->intrblk;
		cpu->intrblk = stat->cpu_sysinfo.intrblk;
		rates->xcalls = (double)(stat->cpu_sysinfo.xcalls - cpu->xcalls) * td;
		Sysstat.XcallsR += rates->xcalls;
		cpu->xcalls = stat->cpu_sysinfo.xcalls;
		rates->pswitch = (double)(stat->cpu_sysinfo.pswitch - cpu->pswitch) * td;
		Sysstat.PswitchR += rates->pswitch;
		cpu->pswitch = stat->cpu_sysinfo.pswitch;
		rates->migrate = (double)(stat->cpu_sysinfo.cpumigrate - cpu->migrate) * td;
		Sysstat.MigrateR += rates->migrate;
		cpu->migrate = stat->cpu_sysinfo.cpumigrate;
		rates->inv_swtch = (double)(stat->cpu_sysinfo.inv_swtch - cpu->inv_swtch) * td;
		Sysstat.InvSwtchR += rates->inv_swtch;
		cpu->inv_swtch = stat->cpu_sysinfo.inv_swtch;
		rates->syscall = (double)(stat->cpu_sysinfo.syscall - cpu->syscall) * td;
		Sysstat.SyscallR += rates->syscall;
		cpu->syscall = stat->cpu_sysinfo.syscall;
		rates->sysfork = (double)(stat->cpu_sysinfo.sysfork - cpu->sysfork) * td;
		cpu->sysfork = stat->cpu_sysinfo.sysfork;
		rates->sysfork += (double)(stat->cpu_sysinfo.sysvfork - cpu->sysvfork) * td;
		Sysstat.SysforkR += rates->sysfork;
		cpu->sysvfork = stat->cpu_sysinfo.sysvfork;
		rates->sysexec = (double)(stat->cpu_sysinfo.sysexec - cpu->sysexec) * td;
		Sysstat.SysexecR += rates->sysexec;
		cpu->sysexec = stat->cpu_sysinfo.sysexec;
		rates->sysread = (double)(stat->cpu_sysinfo.sysread - cpu->sysread) * td;
		Sysstat.SysreadR += rates->sysread;
		cpu->sysread = stat->cpu_sysinfo.sysread;
		rates->syswrite = (double)(stat->cpu_sysinfo.syswrite - cpu->syswrite) * td;
		Sysstat.SyswriteR += rates->syswrite;
		cpu->syswrite = stat->cpu_sysinfo.syswrite;
		rates->anonpgin = (double)(stat->cpu_vminfo.anonpgin - cpu->anonpgin) * td;
		Sysstat.AnonpginR += rates->anonpgin;
		cpu->anonpgin = stat->cpu_vminfo.anonpgin;
		rates->anonpgout = (double)(stat->cpu_vminfo.anonpgout - cpu->anonpgout) * td;
		Sysstat.AnonpgoutR += rates->anonpgout;
		cpu->anonpgout = stat->cpu_vminfo.anonpgout;
		rates->pgswapin = (double)(stat->cpu_vminfo.pgswapin - cpu->pgswapin) * td;
		Sysstat.PgswapinR += rates->pgswapin;
		cpu->pgswapin = stat->cpu_vminfo.pgswapin;
		rates->pgswapout = (double)(stat->cpu_vminfo.pgswapout - cpu->pgswapout) * td;
		Sysstat.PgswapoutR += rates->pgswapout;
		cpu->pgswapout = stat->cpu_vminfo.pgswapout;
		rates->scan = (double)(stat->cpu_vminfo.scan - cpu->scan) * td;
		Sysstat.ScanR += rates->scan;
		cpu->scan = stat->cpu_vminfo.scan;
		rates->anonfree = (double)(stat->cpu_vminfo.anonfree - cpu->anonfree) * td;
		Sysstat.AnonfreeR += rates->anonfree;
		cpu->anonfree = stat->cpu_vminfo.anonfree;
		rates->as_fault = (double)(stat->cpu_vminfo.as_fault - cpu->as_fault) * td;
		Sysstat.AsFaultR += rates->as_fault;
		cpu->as_fault = stat->cpu_vminfo.as_fault;
		rates->hat_fault = (double)(stat->cpu_vminfo.hat_fault - cpu->hat_fault) * td;
		Sysstat.HatFaultR += rates->hat_fault;
		cpu->hat_fault = stat->cpu_vminfo.hat_fault;
		rates->majfault = (double)(stat->cpu_vminfo.maj_fault - cpu->majfault) * td;
		Sysstat.MajFaultR += rates->majfault;
		cpu->majfault = stat->cpu_vminfo.maj_fault;
		rates->rwrf = (double)(stat->cpu_sysinfo.rw_rdfails - cpu->rwrf) * td;
		Sysstat.LockfR += rates->rwrf;
		cpu->rwrf = stat->cpu_sysinfo.rw_rdfails;
		rates->rwwf = (double)(stat->cpu_sysinfo.rw_wrfails - cpu->rwwf) * td;
		Sysstat.LockfR += rates->rwwf;
		cpu->rwwf = stat->cpu_sysinfo.rw_wrfails;
		rates->mxae = (double)(stat->cpu_sysinfo.mutex_adenters - cpu->mxae) * td;
		Sysstat.LockfR += rates->mxae;
		cpu->mxae = stat->cpu_sysinfo.mutex_adenters;
		cpu->timestamp = cpustat->ks_snaptime;

		if (Cpus[inst]->cpusys == 0) {
			int total = stat->cpu_sysinfo.cpu[CPU_USER]
				+ stat->cpu_sysinfo.cpu[CPU_KERNEL]
				+ stat->cpu_sysinfo.cpu[CPU_IDLE]
				+ stat->cpu_sysinfo.cpu[CPU_WAIT]
				- cpu->user
				- cpu->kernel
				- cpu->idle
				- cpu->wait;
			double q = 1 / (double)(total) * 100;
			rates->user = (double)(stat->cpu_sysinfo.cpu[CPU_USER] - cpu->user) * q;
			Sysstat.UserP += rates->user / Sysstat.Ocpus;
			cpu->user = stat->cpu_sysinfo.cpu[CPU_USER];
			rates->kernel = (double)(stat->cpu_sysinfo.cpu[CPU_KERNEL] - cpu->kernel) * q;
			Sysstat.KernelP += rates->kernel / Sysstat.Ocpus;
			cpu->kernel = stat->cpu_sysinfo.cpu[CPU_KERNEL];
			rates->idle = (double)(stat->cpu_sysinfo.cpu[CPU_IDLE] - cpu->idle) * q;
			Sysstat.IdleP += rates->idle / Sysstat.Ocpus;
			cpu->idle = stat->cpu_sysinfo.cpu[CPU_IDLE];
			rates->wait = (double)(stat->cpu_sysinfo.cpu[CPU_WAIT] - cpu->wait) * q;
			Sysstat.IdleP += rates->wait / Sysstat.Ocpus;
			cpu->wait = stat->cpu_sysinfo.cpu[CPU_WAIT];
		}
	}
	if (num == Sysstat.Ocpus)
		return 0;
	assert(num < Sysstat.Ocpus);
	return 1;
}


static int
get_net_stats(void)
{
	Sysstat.NetInR = Sysstat.NetOutR = 0;
	Sysstat.NetInB = Sysstat.NetOutB = 0;
	for (netdata_t *n = Nets; n; n = n->next) {
		kstat_named_t *data;
		if (n->kstat == 0) {
			continue;
		}
		if (kstat_read(KS,n->kstat,0) == -1) {
			/* network interface has been taken offline */
			n->kstat = 0;
			return 1;
		}
		if (n->kstat->ks_type != KSTAT_TYPE_NAMED) {
			continue;
		}
		data = kstat_data_lookup(n->kstat,"ipackets64");
		if (data) {
			assert(data->data_type == KSTAT_DATA_UINT64);
			if (n->value[0]) {
				 n->rates.ipckts = (double)(data->value.ui64 - n->value[0]) / 
					(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
				 Sysstat.NetInR += n->rates.ipckts;
			}
			n->value[0] = data->value.ui64;
			data = kstat_data_lookup(n->kstat,"opackets64");
			assert(data && (data->data_type == KSTAT_DATA_UINT64));
			if (n->value[1]) {
				n->rates.opckts = (double)(data->value.ui64 - n->value[1]) / 
					(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9;
				Sysstat.NetOutR += n->rates.opckts;
			}
			n->value[1] = data->value.ui64;
		}
		data = kstat_data_lookup(n->kstat,"obytes64");
		if (data) {
			assert(data->data_type == KSTAT_DATA_UINT64);
			if (n->value[2]) {
				n->rates.outbs = (double)(data->value.ui64 - n->value[2]) / 
					(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9 / 1024;
				Sysstat.NetOutB += n->rates.outbs;
			}
			n->value[2] = data->value.ui64;
			data = kstat_data_lookup(n->kstat,"rbytes64");
			assert(data && (data->data_type == KSTAT_DATA_UINT64));
			if (n->value[3]) {
				 n->rates.inbs = (double)(data->value.ui64 - n->value[3]) / 
					(double)(n->kstat->ks_snaptime - n->timestamp) * 1E9 / 1024;
				 Sysstat.NetInB += n->rates.inbs;
			}
			n->value[3] = data->value.ui64;
			n->timestamp = n->kstat->ks_snaptime;
		}
	}
	return 0;
}


static void
setup_nets(void)
{
	kstat_t *k;
	netdata_t *n = Nets;
	while (n) {
		n->kstat = 0;
		n->rates.next = 0;
		n = n->next;
	}
	Netrates = 0;
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		size_t l = strlen(k->ks_name) - 1;
		if ((0 == memcmp(k->ks_class,"net",4))
			&& (k->ks_name[l] >= '0') && (k->ks_name[l] <= '9')
			&& (0 != strncmp(k->ks_name,"wrsmd",5))
			&& (-1 != kstat_read(KS,k,0))
			&& (0 != kstat_data_lookup(k,"ipackets64"))) {
			netdata_t *d = Nets;
			while (d && strcmp(d->rates.device,k->ks_name))
				d = d->next;
			if (d == 0) {
				d = calloc(1,sizeof(netdata_t));
				d->next = Nets;
				Nets = d;
				strcpy(d->rates.device,k->ks_name);
				d->rates.inbs = 0;
				d->rates.outbs = 0;
				d->rates.ipckts = 0;
				d->rates.opckts = 0;
			} else {
				assert(0 == strcmp(d->rates.device,k->ks_name));
			}
			d->kstat = k;
			d->rates.next = Netrates;
			Netrates = &d->rates;
		}
	}
}


static void
setup_cpus(void)
{
	unsigned num = 0, maxid = Maxid;
	int i;
	kstat_t *k;

#ifndef _SC_CPUID_MAX
#define _SC_CPUID_MAX 517
#endif
#ifndef NCPU
#define NCPU 256
#endif
	Maxid = sysconf(_SC_CPUID_MAX);
	if (Maxid < 0) {
		warning("unable to determine maximum cpu id\n");
		Maxid = NCPU;
	}
	
	/* number of cpu change during runtime */
	while (maxid > Maxid) {
		free(Cpus[maxid--]);
	}
	Cpus = realloc(Cpus,sizeof(cpudata_t *) * (Maxid + 1));
	assert(Cpus);
	while (maxid <= Maxid) {
		Cpus[maxid] = calloc(1,sizeof(cpudata_t));
		assert(Cpus[maxid]);
		++maxid;
	}
	maxid = Maxid;
	assert(Sysstat.Ncpus != 0);
	for (i = 0; i <= Maxid; ++i) {
		Cpus[i]->cpustat = 0;
		Cpus[i]->cpusys = 0;
	}
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if (0 == strcmp(k->ks_module,"cpu_stat")) {
			assert(k->ks_type == KSTAT_TYPE_RAW);
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpustat = k;
			if (Cpus[k->ks_instance]->rates == 0) {
				Cpus[k->ks_instance]->rates = calloc(1,sizeof(cpurates_t));
				assert(Cpus[k->ks_instance]->rates);
			}
			Cpus[k->ks_instance]->rates->id = k->ks_instance;
			dbug("found CPU with id %d\n",k->ks_instance);
			++num;
		} else if ((0 == strcmp(k->ks_module,"cpu"))
				&& (0 == strcmp(k->ks_name,"sys"))) {
			assert(k->ks_instance <= maxid);
			Cpus[k->ks_instance]->cpusys= k;
			assert(k->ks_type == KSTAT_TYPE_NAMED);
		}
	}
	assert(num <= Sysstat.Ncpus);
	if (Sysstat.Ocpus != num) {
		/* Try to reduce the transient effect of onlining of CPUs on
		 * metrics calculation.
		 */
		Sysstat.UserP /= Sysstat.Ocpus;
		Sysstat.UserP *= num;
		Sysstat.KernelP /= Sysstat.Ocpus;
		Sysstat.KernelP *= num;
		Sysstat.IdleP /= Sysstat.Ocpus;
		Sysstat.IdleP *= num;
		Sysstat.Ocpus = num;
		dbug("Ocpu = %d\n",Sysstat.Ocpus);
	}
}


static void
setup_disks(void)
{
	kstat_t *k;
	diskdata_t *d = Disks;
	while (d) {
		d->kstat = 0;
		d->rates.next = 0;
		d = d->next;
	}
	Diskrates = 0;
	for (k = KS->kc_chain; k != 0; k = k->ks_next) {
		if ((0 == memcmp(k->ks_class,"disk",5))
			&& ((0 == memcmp(k->ks_module,"sd",3))
			|| (0 == memcmp(k->ks_module,"ssd",4))
			|| (0 == memcmp(k->ks_module,"cmdk",5)))) {
			diskdata_t *d = Disks;
			while (d && strcmp(d->rates.device,k->ks_name))
				d = d->next;
			if (d == 0) {
				d = calloc(1,sizeof(diskdata_t));
				d->next = Disks;
				Disks = d;
				strcpy(d->rates.device,k->ks_name);
				d->rates.disk[0] = 0;
				d->rates.major = get_disk_name(k->ks_module,-1,k->ks_instance,d->rates.disk);
			} else {
				assert(0 == strcmp(d->rates.device,k->ks_name));
				if (d->rates.disk[0] == 0)
					d->rates.major = get_disk_name(k->ks_module,d->rates.major,k->ks_instance,d->rates.disk);
			}
			d->kstat = k;
			d->rates.next = Diskrates;
			Diskrates = &d->rates;
		}
	}
}

static int
reconfigure(void)
{
	static kstat_t *sysmisc = 0;
	static kstat_named_t *data = 0;
	int upd;
	kid_t kid;

	upd = kstat_chain_update(KS);
	if ((0 == upd) && Sysstat.Ncpus)
		return 1;
	if (-1 == upd)
		error("Kstat update failed: %s\n",strerror(errno));
	if (sysmisc == 0) {
		sysmisc = kstat_lookup(KS,"unix",0,"system_misc");
		assert(sysmisc && (sysmisc->ks_type == KSTAT_TYPE_NAMED));
	}
	kid = kstat_read(KS,sysmisc,0);
	assert(kid != -1);
	data = kstat_data_lookup(sysmisc,"ncpus");
	assert(data && data->data_type == KSTAT_DATA_UINT32);
	Sysstat.Ncpus = data->value.ui32;
	setup_nets();
	setup_cpus();
	setup_disks();
	Err = get_chain_by_class("device_error",Err);
	Nfs = get_class_chain("nfs","nfs",0,Nfs);
	NfsServer = get_chain_by_name("nfs_server",NfsServer);
	return 1;
}

struct fs_kstats {
	kstat_named_t 
		*read_bytes, *readdir_bytes, *write_bytes, *naccess, 
		*naddmap, *nclose, *ncmp, *ncreate, *ndelmap, *ndispose, 
		*ndump, *ndumpctl, *nfid, *nfrlock, *nfsync, *ngetattr, 
		*ngetpage, *ngetsecattr, *ninactive, *nioctl, *nlink, 
		*nlookup, *nmap, *nmkdir, *nopen, *npageio, *npathconf, 
		*npoll, *nputpage, *nread, *nreaddir, *nreadlink, *nrealvp, 
		*nremove, *nrename, *nrmdir, *nrwlock, *nrwunlock, *nseek, 
		*nsetattr, *nsetfl, *nsetsecattr, *nshrlock, *nspace, *nsymlink, 
		*nvnevent, *nwrite; 
};

static void
lookup_fs_stats(kstat_t *ks, struct fs_kstats *fs)
{
	int i = ks->ks_ndata;
	while (i--) {
		kstat_named_t *s = &((kstat_named_t *) ks->ks_data)[i];
		switch (s->name[0]) {
		case 'n':
			switch (s->name[1]) {
			case 'a':
				if (strcmp("naccess"+2,s->name+2) == 0)
					fs->naccess = s;
				else if (strcmp("naddmap"+2,s->name+2) == 0)
					fs->naddmap = s;
				break;
			case 'c':
				if (strcmp("nclose"+2,s->name+2) == 0)
					fs->nclose = s;
				else if (strcmp("ncmp"+2,s->name+2) == 0)
					fs->ncmp = s;
				else if (strcmp("ncreate"+2,s->name+2) == 0)
					fs->ncreate = s;
				break;
			case 'd':
				if (strcmp("ndelmap"+2,s->name+2) == 0)
					fs->ndelmap = s;
				else if (strcmp("ndispose"+2,s->name+2) == 0)
					fs->ndispose = s;
				else if (strcmp("ndump"+2,s->name+2) == 0)
					fs->ndump = s;
				else if (strcmp("ndumpctl"+2,s->name+2) == 0)
					fs->ndumpctl = s;
				break;
			case 'f':
				if (strcmp("nfid"+2,s->name+2) == 0)
					fs->nfid = s;
				else if (strcmp("nfrlock"+2,s->name+2) == 0)
					fs->nfrlock = s;
				else if (strcmp("nfsync"+2,s->name+2) == 0)
					fs->nfsync = s;
				break;
			case 'g':
				if (strcmp("ngetattr"+2,s->name+2) == 0)
					fs->ngetattr = s;
				else if (strcmp("ngetpage"+2,s->name+2) == 0)
					fs->ngetpage = s;
				else if (strcmp("ngetsecattr"+2,s->name+2) == 0)
					fs->ngetsecattr = s;
				break;
			case 'i':
				if (strcmp("ninactive"+2,s->name+2) == 0)
					fs->ninactive = s;
				else if (strcmp("nioctl"+2,s->name+2) == 0)
					fs->nioctl = s;
				break;
			case 'l':
				if (strcmp("nlink"+2,s->name+2) == 0)
					fs->nlink = s;
				else if (strcmp("nlookup"+2,s->name+2) == 0)
					fs->nlookup = s;
				break;
			case 'm':
				if (strcmp("nmap"+2,s->name+2) == 0)
					fs->nmap = s;
				else if (strcmp("nmkdir"+2,s->name+2) == 0)
					fs->nmkdir = s;
				break;
			case 'o':
				if (strcmp("nopen"+2,s->name+2) == 0)
					fs->nopen = s;
				break;
			case 'p':
				if (strcmp("npageio"+2,s->name+2) == 0)
					fs->npageio = s;
				else if (strcmp("npathconf"+2,s->name+2) == 0)
					fs->npathconf = s;
				else if (strcmp("npoll"+2,s->name+2) == 0)
					fs->npoll = s;
				else if (strcmp("nputpage"+2,s->name+2) == 0)
					fs->nputpage = s;
				break;
			case 'r':
				if (strcmp("nread"+2,s->name+2) == 0)
					fs->nread = s;
				else if (strcmp("nreaddir"+2,s->name+2) == 0)
					fs->nreaddir = s;
				else if (strcmp("nreadlink"+2,s->name+2) == 0)
					fs->nreadlink = s;
				else if (strcmp("nrealvp"+2,s->name+2) == 0)
					fs->nrealvp = s;
				else if (strcmp("nremove"+2,s->name+2) == 0)
					fs->nremove = s;
				else if (strcmp("nrename"+2,s->name+2) == 0)
					fs->nrename = s;
				else if (strcmp("nrmdir"+2,s->name+2) == 0)
					fs->nrmdir = s;
				else if (strcmp("nrwlock"+2,s->name+2) == 0)
					fs->nrwlock = s;
				else if (strcmp("nrwunlock"+2,s->name+2) == 0)
					fs->nrwunlock = s;
				break;
			case 's':
				if (strcmp("nseek"+2,s->name+2) == 0)
					fs->nseek = s;
				else if (strcmp("nsetattr"+2,s->name+2) == 0)
					fs->nsetattr = s;
				else if (strcmp("nsetfl"+2,s->name+2) == 0)
					fs->nsetfl = s;
				else if (strcmp("nsetsecattr"+2,s->name+2) == 0)
					fs->nsetsecattr = s;
				else if (strcmp("nshrlock"+2,s->name+2) == 0)
					fs->nshrlock = s;
				else if (strcmp("nspace"+2,s->name+2) == 0)
					fs->nspace = s;
				else if (strcmp("nsymlink"+2,s->name+2) == 0)
					fs->nsymlink = s;
				break;
			case 'v':
				if (strcmp("nvnevent"+2,s->name+2) == 0)
					fs->nvnevent = s;
				break;
			case 'w':
				if (strcmp("nwrite"+2,s->name+2) == 0)
					fs->nwrite = s;
				break;
			default:
				break;
			}
			break;
		case 'r':
			if (strcmp("read_bytes"+1,s->name+1) == 0)
				fs->read_bytes = s;
			else if (strcmp("readdir_bytes"+1,s->name+1) == 0)
				fs->readdir_bytes = s;
			break;
		case 'w':
			if (strcmp("write_bytes"+1,s->name+1) == 0)
				fs->write_bytes = s;
			break;
		default:
			break;
		}
	}
}


static int
update_fs_rates(void)
{
	static class_chain_t *fs_chain = 0;
	class_chain_t *c;
	fsstat_t *fs = get_fs_stats();
	fs_chain = get_chain_by_name("vopstats_",fs_chain);
	c = fs_chain;
	while (c) {
		unsigned long nodeid;
		kstat_t *s = c->kstat;
		kstat_named_t *data;
		fsstat_t *f = fs;
		float td;
		unsigned long long ops = 0;
		struct fs_kstats fsks;

		if ((c->iteration == Iteration) || (s == 0)) {
			c = c->next;
			continue;
		}
		if (1 != sscanf(s->ks_name+9,"%lx",&nodeid)) {
			c = c->next;
			continue;
		}
		dbug("nodeid, %x\n",nodeid);
		while (f && (f->fsid != nodeid))
			f = f->next;
		if (f == 0) {
			c = c->next;
			continue;
		}
		dbug("vopstat: %s, nodeid, %x, fs: %s\n",s->ks_name,nodeid,f->dev);
		if (-1 == kstat_read(KS,s,0))
			return 1;
		assert((s->ks_type == KSTAT_TYPE_NAMED)
			&& (s->ks_snaptime != c->timestamp));
		bzero(&fsks,sizeof(fsks));
		lookup_fs_stats(s,&fsks); 

		td = 1E9 / (float)(s->ks_snaptime - f->rate_ts);
		data = fsks.read_bytes;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->read_rate = (float)((data->value.ui64 - f->numread) >> 10) * td;
		dbug("read: %llu-%llu/%llu-%llu, %g\n",data->value.ui64,f->numread,s->ks_snaptime,f->rate_ts,f->read_rate);
		f->numread = data->value.ui64;

		data = fsks.readdir_bytes;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->read_rate += (float)((data->value.ui64 - f->numreaddir) >> 10) * td; 
		//dbug("readdir: %llu-%llu/%llu-%llu, %g\n",data->value.ui64,f->numread,s->ks_snaptime,f->rate_ts,f->read_rate);
		f->numreaddir = data->value.ui64;

		data = fsks.write_bytes;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		f->write_rate = (float)((data->value.ui64 - f->numwrite) >> 10) * td;
		dbug("write: %llu-%llu, %g\n",data->value.ui64,f->numwrite,f->write_rate);
		f->numwrite = data->value.ui64;

		data = fsks.naccess;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->naccess)
			f->naccess = data->value.ui64;
		ops += data->value.ui64 - f->naccess;
		//dbug("naccess = %llu\n",data->value.ui64-f->naccess);
		f->naccess = data->value.ui64;

		data = fsks.naddmap;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->naddmap)
			f->naddmap = data->value.ui64;
		ops += data->value.ui64 - f->naddmap;
		//dbug("naddmap = %llu\n",data->value.ui64-f->naddmap);
		f->naddmap = data->value.ui64;

		data = fsks.nclose;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nclose)
			f->nclose = data->value.ui64;
		ops += data->value.ui64 - f->nclose;
		//dbug("nclose = %llu\n",data->value.ui64-f->nclose);
		f->nclose = data->value.ui64;

		data = fsks.ncmp;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ncmp;
		//dbug("ncmp = %llu\n",data->value.ui64-f->ncmp);
		f->ncmp = data->value.ui64;

		data = fsks.ncreate;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ncreate;
		//dbug("ncreate = %llu\n",data->value.ui64-f->ncreate);
		f->ncreate = data->value.ui64;

		data = fsks.ndelmap;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndelmap;
		//dbug("ndelmap = %llu\n",data->value.ui64-f->ndelmap);
		f->ndelmap = data->value.ui64;

		data = fsks.ndispose;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndispose;
		//dbug("ndispose = %llu\n",data->value.ui64-f->ndispose);
		f->ndispose = data->value.ui64;
		
		data = fsks.ndump;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndump;
		//dbug("ndump = %llu\n",data->value.ui64-f->ndump);
		f->ndump = data->value.ui64;

		data = fsks.ndumpctl;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ndumpctl;
		//dbug("ndumpctl = %llu\n",data->value.ui64-f->ndumpctl);
		f->ndumpctl = data->value.ui64;

		data = fsks.nfid;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfid;
		//dbug("nfid = %llu\n",data->value.ui64-f->nfid);
		f->nfid = data->value.ui64;

		data = fsks.nfrlock;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfrlock;
		//dbug("nfrlock = %llu\n",data->value.ui64-f->nfrlock);
		f->nfrlock = data->value.ui64;

		data = fsks.nfsync;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->nfsync;
		//dbug("nfsync = %llu\n",data->value.ui64-f->nfsync);
		f->nfsync = data->value.ui64;

		data = fsks.ngetattr;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ngetattr;
		//dbug("ngetattr = %llu\n",data->value.ui64-f->ngetattr);
		f->ngetattr = data->value.ui64;

		data = fsks.ngetpage;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		ops += data->value.ui64 - f->ngetpage;
		//dbug("ngetpage = %llu\n",data->value.ui64-f->ngetpage);
		f->ngetpage = data->value.ui64;

		data = fsks.ngetsecattr;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->ngetsecattr)
			f->ngetsecattr = data->value.ui64;
		ops += data->value.ui64 - f->ngetsecattr;
		//dbug("ngetsecattr = %llu\n",data->value.ui64-f->ngetsecattr);
		f->ngetsecattr = data->value.ui64;

		data = fsks.ninactive;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->ninactive)
			f->ninactive = data->value.ui64;
		ops += data->value.ui64 - f->ninactive;
		//dbug("ninactive = %llu\n",data->value.ui64-f->ninactive);
		f->ninactive = data->value.ui64;

		data = fsks.nioctl;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nioctl)
			f->nioctl = data->value.ui64;
		ops += data->value.ui64 - f->nioctl;
		//dbug("nioctl = %llu\n",data->value.ui64-f->nioctl);
		f->nioctl = data->value.ui64;

		data = fsks.nlink;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nlink)
			f->nlink = data->value.ui64;
		ops += data->value.ui64 - f->nlink;
		//dbug("nlink = %llu\n",data->value.ui64-f->nlink);
		f->nlink = data->value.ui64;

		data = fsks.nlookup;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nlookup)
			f->nlookup = data->value.ui64;
		ops += data->value.ui64 - f->nlookup;
		//dbug("nlookup = %llu\n",data->value.ui64-f->nlookup);
		f->nlookup = data->value.ui64;

		data = fsks.nmap;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nmap)
			f->nmap = data->value.ui64;
		ops += data->value.ui64 - f->nmap;
		//dbug("nmap = %llu\n",data->value.ui64-f->nmap);
		f->nmap = data->value.ui64;

		data = fsks.nmkdir;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nmkdir)
			f->nmkdir = data->value.ui64;
		ops += data->value.ui64 - f->nmkdir;
		//dbug("nmkdir = %llu\n",data->value.ui64-f->nmkdir);
		f->nmkdir = data->value.ui64;

		data = fsks.nopen;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nopen)
			f->nopen = data->value.ui64;
		ops += data->value.ui64 - f->nopen;
		//dbug("nopen = %llu\n",data->value.ui64-f->nopen);
		f->nopen = data->value.ui64;

		data = fsks.npageio;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->npageio)
			f->npageio = data->value.ui64;
		ops += data->value.ui64 - f->npageio;
		//dbug("npageio = %llu\n",data->value.ui64-f->npageio);
		f->npageio = data->value.ui64;

		data = fsks.npathconf;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->npathconf)
			f->npathconf = data->value.ui64;
		ops += data->value.ui64 - f->npathconf;
		//dbug("npathconf = %llu\n",data->value.ui64-f->npathconf);
		f->npathconf = data->value.ui64;

		data = fsks.npoll;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->npoll)
			f->npoll = data->value.ui64;
		ops += data->value.ui64 - f->npoll;
		//dbug("npoll = %llu\n",data->value.ui64-f->npoll);
		f->npoll = data->value.ui64;

		data = fsks.nputpage;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nputpage)
			f->nputpage = data->value.ui64;
		ops += data->value.ui64 - f->nputpage;
		//dbug("nputpage = %llu\n",data->value.ui64-f->nputpage);
		f->nputpage = data->value.ui64;

		data = fsks.nread;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nread)
			f->nread = data->value.ui64;
		ops += data->value.ui64 - f->nread;
		//dbug("nread = %llu\n",data->value.ui64-f->nread);
		f->nread = data->value.ui64;

		data = fsks.nreaddir;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nreaddir)
			f->nreaddir = data->value.ui64;
		ops += data->value.ui64 - f->nreaddir;
		//dbug("nreaddir = %llu\n",data->value.ui64-f->nreaddir);
		f->nreaddir = data->value.ui64;

		data = fsks.nreadlink;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nreadlink)
			f->nreadlink = data->value.ui64;
		ops += data->value.ui64 - f->nreadlink;
		//dbug("nreadlink = %llu\n",data->value.ui64-f->nreadlink);
		f->nreadlink = data->value.ui64;

		data = fsks.nrealvp;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nrealvp)
			f->nrealvp = data->value.ui64;
		ops += data->value.ui64 - f->nrealvp;
		//dbug("nrealvp = %llu\n",data->value.ui64-f->nrealvp);
		f->nrealvp = data->value.ui64;

		data = fsks.nremove;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nremove)
			f->nremove = data->value.ui64;
		ops += data->value.ui64 - f->nremove;
		//dbug("nremove = %llu\n",data->value.ui64-f->nremove);
		f->nremove = data->value.ui64;

		data = fsks.nrename;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nrename)
			f->nrename = data->value.ui64;
		ops += data->value.ui64 - f->nrename;
		//dbug("nrename = %llu\n",data->value.ui64-f->nrename);
		f->nrename = data->value.ui64;

		data = fsks.nrmdir;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nrmdir)
			f->nrmdir = data->value.ui64;
		ops += data->value.ui64 - f->nrmdir;
		//dbug("nrmdir = %llu\n",data->value.ui64-f->nrmdir);
		f->nrmdir = data->value.ui64;

		data = fsks.nrwlock;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nrwlock)
			f->nrwlock = data->value.ui64;
		ops += data->value.ui64 - f->nrwlock;
		//dbug("nrwlock = %llu\n",data->value.ui64-f->nrwlock);
		f->nrwlock = data->value.ui64;

		data = fsks.nrwunlock;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nrwunlock)
			f->nrwunlock = data->value.ui64;
		ops += data->value.ui64 - f->nrwunlock;
		//dbug("nrwunlock = %llu\n",data->value.ui64-f->nrwunlock);
		f->nrwunlock = data->value.ui64;

		data = fsks.nseek;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nseek)
			f->nseek = data->value.ui64;
		ops += data->value.ui64 - f->nseek;
		//dbug("nseek = %llu\n",data->value.ui64-f->nseek);
		f->nseek = data->value.ui64;

		data = fsks.nsetattr;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nsetattr)
			f->nsetattr = data->value.ui64;
		ops += data->value.ui64 - f->nsetattr;
		//dbug("nsetattr = %llu\n",data->value.ui64-f->nsetattr);
		f->nsetattr = data->value.ui64;

		data = fsks.nsetfl;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nsetfl)
			f->nsetfl = data->value.ui64;
		ops += data->value.ui64 - f->nsetfl;
		//dbug("nsetfl = %llu\n",data->value.ui64-f->nsetfl);
		f->nsetfl = data->value.ui64;

		data = fsks.nsetsecattr;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nsetsecattr)
			f->nsetsecattr = data->value.ui64;
		ops += data->value.ui64 - f->nsetsecattr;
		//dbug("nsetsecattr = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nsetsecattr = data->value.ui64;

		data = fsks.nshrlock;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nshrlock)
			f->nshrlock = data->value.ui64;
		ops += data->value.ui64 - f->nshrlock;
		//dbug("nshrlock = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nshrlock = data->value.ui64;

		data = fsks.nspace;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nspace)
			f->nspace = data->value.ui64;
		ops += data->value.ui64 - f->nspace;
		//dbug("nspace = %llu\n",data->value.ui64-f->nsetsecattr);
		f->nspace = data->value.ui64;

		data = fsks.nsymlink;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nsymlink)
			f->nsymlink = data->value.ui64;
		ops += data->value.ui64 - f->nsymlink;
		//dbug("nsymlink = %llu\n",data->value.ui64-f->nsymlink);
		f->nsymlink = data->value.ui64;

		data = fsks.nvnevent;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nvnevent)
			f->nvnevent = data->value.ui64;
		ops += data->value.ui64 - f->nvnevent;
		//dbug("nvnevent = %llu\n",data->value.ui64-f->nvnevent);
		f->nvnevent = data->value.ui64;

		data = fsks.nwrite;
		assert(data && (data->data_type == KSTAT_DATA_UINT64));
		if (data->value.ui64 < f->nwrite)
			f->nwrite = data->value.ui64;
		ops += data->value.ui64 - f->nwrite;
		//dbug("nwrite = %llu\n",data->value.ui64-f->nwrite);
		f->nwrite = data->value.ui64;

		f->ops_rate = (float)(ops) * td;
		dbug("%s: oprate %g, ntotal %llu, ops %llu, dt %lld\n",f->dev,f->ops_rate,f->ntotal,ops,s->ks_snaptime-f->rate_ts);
		f->ntotal = ops;

		f->rate_ts = s->ks_snaptime;
		c->iteration = Iteration;
		dbug("%s: I/O(%g/%g), %lld %lld\n",f->dev,f->read_rate,f->write_rate,c->value[0],c->value[1]);
		c = c->next;
	}
	while (fs) {
		if (fs->zroot) {
			fsstat_t *z = fs->zroot;
			dbug("accounting %s in %s\n",fs->dev,z->dev);
			z->write_rate += fs->write_rate; 
			z->read_rate += fs->read_rate; 
			z->ops_rate += fs->ops_rate; 
		}
		fs = fs->next;
	}
	return 0;
}


static void
get_local_data(int userdata, int fsdata)
{
	bzero(&Sysstat.TS,(char *)&Sysstat.cpurates-(char *)&Sysstat.TS);
	Cpurates = 0;
	while (reconfigure()) {
		if (get_cpu_stat())
			continue;
		if (Diskstats && get_disk_io())
			continue;
		if (fsdata && update_fs_stats())
			continue;
		if (fsdata && update_fs_rates())
			continue;
		if (get_nfs_io())
			continue;
		if (get_unix_systempages())
			continue;
		if (get_zfs_arc())
			continue;
		if (Ipstat) {
			/* Solaris 10 extras */
			if (get_cpu_sys())
				continue;
			get_ip_ip();
		}
		if (get_net_stats())
			continue;
		if (Diskstats)
			get_device_errors();
		get_swapinfo();
		++Iteration;
		Sysstat.TS = time(0);
		getloadavg(Sysstat.loadavg,3);
		Sysstat.fs = get_fs_stats();
		if (userdata && Userstats)
			Sysstat.us = get_procinfo();
		Sysstat.cpurates = Cpurates;
		Sysstat.diskrates = Diskrates;
		Sysstat.netrates = Netrates;
		Sysstat.displayed = 0;
		return;
	}
}


static void
init_localstat(sysstat_t *s)
{
	int ret;
	char *next;
	struct rlimit rlimit;

	if (s == 0)
		s = &Sysstat;
	if (-1 == getrlimit(RLIMIT_NOFILE,&rlimit))
		info("could not get resource limit: %s\n",strerror(errno));
	rlimit.rlim_cur = rlimit.rlim_max;
	if (-1 == setrlimit(RLIMIT_NOFILE,&rlimit))
		warning("could not reset no of files limit: %s\n",strerror(errno));
	get_boottime();
	KS = kstat_open();
	if (KS == 0)
		error("cannot access kernel statistics: %s\n",strerror(errno));
	Syspages = kstat_lookup(KS,"unix",0,"system_pages");
	assert(Syspages && (Syspages->ks_type == KSTAT_TYPE_NAMED));
	ZfsArc = kstat_lookup(KS,"zfs",0,"arcstats");
	Ipstat = kstat_lookup(KS,"ip",0,"ip");
	setup_nets();
	if (Ipstat == 0) {
		/* before Solaris 10 */
		Sysstat.IpInR = Sysstat.IpOutR = -1;
	} else 
		assert(Ipstat->ks_type == KSTAT_TYPE_NAMED);
	Pgsz = getpagesize();
	ret = gethostname(s->Hostinfo,sizeof(s->Hostinfo));
	assert(ret == 0);
	if ((next = strchr(s->Hostinfo,'.')) == 0)
		next = s->Hostinfo + strlen(s->Hostinfo);
	*next = ' ';
	++next;
	sysinfo(SI_SYSNAME,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_RELEASE,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	next += strlen(next);
	*next = ' ';
	++next;
	sysinfo(SI_PLATFORM,next,sizeof(s->Hostinfo) - (next - s->Hostinfo));
	reconfigure();
	/* getdata once then sleep 0.3sec to get reasonable start values */
	get_local_data(1,1);
	(void) usleep(300000);	
}


static void
daemon_sighandler(int s)
{
	assert(s == SIGTERM || s == SIGQUIT || s == SIGINT || s == SIGPIPE);
	if (Localmode == 0)
		(void) unlink("/tmp/sysstat.pid");
	exit(EXIT_SUCCESS);
}


void
rundaemon(unsigned interval, unsigned fsrate, unsigned userrate, unsigned count)
{
	int c = 0;
	unsigned fsr = fsrate, usr = userrate;
	
	gather_disk_infos();
	(void) signal(SIGTERM,daemon_sighandler);
	(void) signal(SIGQUIT,daemon_sighandler);
	(void) signal(SIGINT,daemon_sighandler);
	if (Localmode)
		(void) signal(SIGPIPE,daemon_sighandler);
	init_zfs();
	init_userstat();
	init_localstat(&Sysstat);
	for (;(count == 0) || (c != count); (void)(count ? ++c : c)) {
		--fsr;
		--usr;
		get_local_data(usr == 0,fsr == 0);
		if (fsr == 0)
			fsr = fsrate;
		if (usr == 0)
			usr = userrate;
		send_to_clients(&Sysstat,Localmode);
		(void) sleep(interval);
	}
}

// vim:tw=0
