/*
 * (c) 2006-2009 Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include <assert.h>
#include <curses.h>
#include <errno.h>
#include <fcntl.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <strings.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/utsname.h>
#include <unistd.h>
#include "log.h"
#include "sysstat.h"

#ifdef __GNUC__
extern FILE *popen(const char *,const char *);
extern int pclose(FILE *);
extern int snprintf(char *,size_t,const char *,...);
extern int vsnprintf(char *,size_t,const char *,va_list);
#endif

char Execname[PATH_MAX];
char **ARGV;
int ARGC;
void (*DfltSegV)(int signr) = 0;
static int Syslog = 0, Verbose = 0;

static int Log = -1;


#ifdef DEBUG
void dbug(const char *msg, ...)
{
	va_list val;
	char buf[8192];
	size_t l;

	if (Log == -1)
		return;
	va_start(val,msg);
	l = vsnprintf(buf,sizeof(buf),msg,val);
	va_end(val);
	if (Syslog) {
		syslog(LOG_DEBUG,buf);
	}
	(void) write(Log,buf,l);
}
#endif


void info(const char *msg, ...)
{
	va_list val;
	char buf[8192];
	size_t l;

	if (Verbose == 0)
		return;
	va_start(val,msg);
	l = vsnprintf(buf,sizeof(buf),msg,val);
	va_end(val);
	if (Syslog)
		syslog(LOG_INFO,buf);
	else if (Log != -1)
		(void) write(Log,buf,l);
	else
		(void) write(STDERR_FILENO,buf,strlen(buf));
}


void warning(const char *msg, ...)
{
	va_list val;
	char buf[8192] = "warning: ";
	size_t l = 9;

	va_start(val,msg);
	l += vsnprintf(buf+9,sizeof(buf)-9,msg,val);
	va_end(val);
	if (Syslog)
		syslog(LOG_WARNING,buf+9);
	else if (Log != -1)
		(void) write(Log,buf,l);
	else
		(void) write(STDERR_FILENO,buf,l);
}


void error(const char *msg, ...)
{
	va_list val;
	char buf[8192] = "error: ";
	size_t l = 7;

	va_start(val,msg);
	l += vsnprintf(buf+7,sizeof(buf)-7,msg,val);
	va_end(val);
	if (Syslog)
		syslog(LOG_ERR,buf+7);
	else if (Log != -1)
		(void) write(Log,buf,l);
	else
		(void) write(STDERR_FILENO,buf,l);
	(void) close(STDERR_FILENO);
	exit(EXIT_FAILURE);
}


void log2syslog(const char *exename)
{
	openlog(exename,LOG_USER,0);
	Syslog = 1;
}


void setVerbose(const char *logfile)
{
	Verbose = 1;
	if (Log == -1) {
		int log = open(logfile,O_WRONLY|O_APPEND|O_CREAT|O_TRUNC,0666);
		if (log == -1)
			warning("could not create %s: %s\n",logfile,strerror(errno));
		else 
			Log = log;
	}
}


void setLog(int l)
{
	Log = l;
}


void SegVhandler(int signr)
{
	struct utsname utsname;
	char msg[16<<10], *m = msg;
	char cmd[32],pwd[PATH_MAX];
	int tty = open("/dev/tty",O_RDWR), i = 0, s = 0;
	pid_t pid = getpid();
	FILE *ps;

	assert(signr == SIGSEGV || signr == SIGBUS);
	if (tty == -1)
		tty = STDERR_FILENO;
	uname(&utsname);
	m += snprintf(msg,sizeof(msg),
		"\rSegment Violation occured. This is probably a bug.\n\r"
		"Please send the following information and files to software@maier-komor.de:\n\r"
		"sysstat version " VERSION "\n\r"
		"sysname   : %s\n\r"
		"release   : %s\n\r"
		"version   : %s\n\r"
		"machine   : %s\n\r"
		"executable: %s\n\r"
		"core file : %s/core.%d (default)\n\r"
		, utsname.sysname
		, utsname.release
		, utsname.version
		, utsname.machine
		, Execname
		, getcwd(pwd,PATH_MAX)
		, (int)pid
		);
	while (i < ARGC) {
		m += sprintf(m,"argv[%d]   : %s\n\r",i,ARGV[i]);
		++i;
	}
	m += sprintf(m, "SegV occured at:\n\r");
#ifdef __sun
	(void) seteuid(0);
	(void) setuid(0);
	sprintf(cmd,"/usr/bin/pstack %d",(int)pid);
	ps = popen(cmd,"r");
	while (!feof(ps)) {
		char buf[1024];
		fgets(buf,sizeof(buf),ps);
		if (s) {
			strcat(m,buf);
			strcat(m,"\r");
		} else if (strstr(buf,"SegVhandler"))
			s = 1;
	}
	pclose(ps);
#else
	m += sprintf(m, "stacktrace unavailable - please analyze core file\n\r");
#endif
#ifdef DEBUG
	if (Log != STDERR_FILENO)
		(void) write(Log,msg,strlen(msg));
#endif
	(void) write(tty,msg,strlen(msg));
	if (DfltSegV) {
		DfltSegV(signr);
	} else {
		char cmd[256];
		sprintf(cmd,"gcore %d",(int)pid);
		(void)system(cmd);
	}
	_exit(EXIT_FAILURE);
}


