/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"

#include <arpa/inet.h>
#include <assert.h>
#include <errno.h>
#include <math.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>


int
	Port = 7536;
char *
	McastAddr = "224.0.0.75";
extern int
	VerboseZFS;

/* The packet size should have been 1472 (equally to MTU in
 * ethernet), but a typo introduced this wrong size initially.
 * Now we are bound to it, to be backward compatible...
 */
#define PACKETSIZE 4173

typedef struct msg_s {
	struct msg_s *next;
	char text[PACKETSIZE];	/* maximum size of one datagram packet plus
				 * terminating \0
				 */
} msg_t;

static char
	*M = 0;
static size_t
	S;
static msg_t
	*Msg = 0, *Cache = 0;
static int
	Sock = -1;
static struct sockaddr_in
	MA;
static sysstat_t *
	SS;

static inline void
save(void)
{
	msg_t *m;

	dbug("save(): %s\n",Msg->text);
	*M = 0;
	if (Cache) {
		m = Cache;
		Cache = m->next;
	} else {
		m = malloc(sizeof(msg_t));
		assert(m);
	}
	m->text[0] = 0;
	m->next = Msg;
	Msg = m;
	M = Msg->text;
	S = sizeof(Msg->text)/sizeof(char) - sizeof(char);
	M += sprintf(M,"sysstat\t1\nHost\t%s\nOCPU\t%ul\n",SS->Hostinfo,SS->Ocpus);
	//add_str("sysstat","1",'\n');
	//add_str("Host",SS->Hostinfo,'\n');
	/* OCPU is needed because of an assertion in terminal.c in older
	 * versions of sysstat. */
	//add_int("OCPU",SS->Ocpus,'\n');
}


static inline char *
per2str(double d)
{
	static char str[7];
	/* allow little rounding errors */
	/* this can possibly occure when summing up multiple user CPU
	 * percentages in the process statistics
	 */
	if (d < 0) {
		sprintf(str,"%.0f",d);
	} else if (d < 2E-2) {
		sprintf(str,"%.0f",d);
	} else if (d < 9.995) {
		sprintf(str,"%.2f",d);
	} else if (d < 99.95) {
		sprintf(str,"%.1f",d);
	} else {
		/* account everything above 100 as 100 */
		//strcpy(str,"100");
		sprintf(str,"%.0f",d);
	}
	return str;
}


static void
add_int(const char *f, long long v, char c)
{
	size_t l;
	char *m;

	assert(f);
	if (S < 4)
start:
		save();
	m = M;
	l = strlcpy(m,f,S);
	if (l+3 >= S) {
		goto start;
	}
	S -= l + 1;
	m += l;
	*m++ = '\t';
	/* c1 = snprintf(m,S,"%lld",v); {*/
	{
		long long d = 10;
		while (d <= v) {
			d *= 10;
		}
		while (d /= 10) {
			long long r;
			r = v / d;
			v %= d;
			*m++ = r + '0';
			if (--S == 0) {
				goto start;
			}
		}
	}
	--S;
	/* c1 = snprintf(m,S,"%lld",v); } */
	*m++ = c;
	M = m;
}


static inline size_t
dbl2str(char *str, float d, size_t n)
{
	const char *f = "%.0f";
	if (d < 0) {
		/*sprintf(str,"%.0f",r);*/
		(void) d;	/* make lint happy */
	} else if (d < 2E-2) {
		/*strcpy(str,"0");*/
		d = 0;
	//} else if (rint(d*100) < 1E3) {
	} else if (d*100 < 1E3) {
		/*sprintf(str,"%.2f",d);*/
		f = "%.2f";
	//} if (rint(d*10) < 1E3) {
	} if (d*10 < 1E3) {
		/*sprintf(str,"%.1f",d);*/
		f = "%.1f";
	} else {
		/*sprintf(str,"%.0f",r);*/
		d = rint(d);
	}
	return snprintf(str,n,f,d);
}


static inline void
add_dbl(char *f, double d, char c)
{
	size_t l;
	char *m;

	assert(f);
#ifdef DEBUG
	if (d > 1E10) {
		warning("value out of range: %s: %g\n",f,d);
	}
#endif
	if (S < 4)
start:
		save();
	m = M;
	l = strlcpy(m,f,S);
	if (l+3 >= S) {
		goto start;
	}
	S -= l;
	m += l;
	*m++ = '\t';
	--S;
	l = dbl2str(m,d,S);
	if (l+1 > S) {
		goto start;
	}
	S -= l;
	m += l;
	*m++ = c;
	--S;
	M = m;
}


static void
add_str(char *restrict f, char *restrict v, char c)
{
	size_t l;
	char *m;

	assert(f && v);
	if (S < 4)
start:
		save();
	m = M;
	l = strlcpy(m,f,S);
	if (l+3 >= S) {
		goto start;
	}
	S -= l;
	m += l;
	*m++ = '\t';
	--S;
	l = strlcpy(m,v,S);
	if (l+1 > S) {
		goto start;
	}
	S -= l;
	m += l;
	*m++ = c;
	--S;
	M = m;
}


static void
sendall(int localmode)
{
	msg_t *m = Msg, *p = 0;
	dbug("sendall()\n");
	assert(m);
	do {
		dbug("send: \n%s\n",m->text);
		if (localmode) {
			if (-1 == write(STDOUT_FILENO,m->text,PACKETSIZE)) {
				if (errno == EPIPE)
					exit(EXIT_SUCCESS);
				error("could not transmit status to client: %s\n",strerror(errno));
			}
		} else {
					
			if (-1 == sendto(Sock, m->text, PACKETSIZE, 0, (const struct sockaddr *) &MA, sizeof(MA))) {
				warning("could not multicast message: %s\n",strerror(errno));
			}
		}
		p = m;
		m = m->next;
		p->text[0] = 0;
		p->next = Cache;
		Cache = p;
	} while (m);
	Msg = Cache;
	Cache = Msg->next;
	Msg->next = 0;
	M = Msg->text;
	S = sizeof(Msg->text)/sizeof(char) - sizeof(char);
	add_str("sysstat","1",'\n');
	add_str("Host",SS->Hostinfo,'\n');
}


void send_to_clients(sysstat_t *s, int localmode)
{
	fsstat_t *fs = s->fs;
	userstat_t *us = s->us;
	cpurates_t *r = s->cpurates;
	diskrates_t *dr = s->diskrates;
	netrates_t *nr = s->netrates;

	dbug("send_to_network()\n");
	if (-1 == Sock) {
		if (localmode) {
			Sock = fileno(stdout);
		} else {
			Sock = socket(AF_INET, SOCK_DGRAM, 0);
			if (-1 == Sock)
				error("could not create socket: %s\n",strerror(errno));
			bzero(&MA,sizeof(MA));
			MA.sin_family = AF_INET;
			MA.sin_port = htons(Port);
#if SOLARIS >= 8
			if (-1 == inet_pton(AF_INET, McastAddr, &MA.sin_addr))
				error("could not set multicast address: %s\n",strerror(errno));
#else
			MA.sin_addr.s_addr = inet_addr(McastAddr);
			if ((in_addr_t)(-1) == MA.sin_addr.s_addr)
				error("could not set multicast address\n");
#endif
		}
		Msg = malloc(sizeof(msg_t));
		Msg->next = 0;
		M = Msg->text;
		S = sizeof(Msg->text)/sizeof(char) - sizeof(char);
		add_str("sysstat","1",'\n');
		add_str("Host",s->Hostinfo,'\n');
	}
	SS = s;

	assert(Msg && Msg->next == 0);
	add_int("TS",s->TS,'\n');
	add_int("BT",s->BootTime,'\n');
	add_int("Itv",Interval,'\n');
	add_int("NCPU",s->Ncpus,'\n');
	add_int("OCPU",s->Ocpus,'\n');
	add_int("Disk",s->Devices,'\n');
	add_str("User",per2str(s->UserP),'\n');
	add_str("Kernel",per2str(s->KernelP),'\n');
	add_str("Idle",per2str(s->IdleP),'\n');
	add_dbl("Load1",s->loadavg[0],'\n');
	add_dbl("Load5",s->loadavg[1],'\n');
	add_dbl("Load15",s->loadavg[2],'\n');
	add_dbl("intr",s->IntrR,'\n');
	add_dbl("intrblk",s->IntrBlkR,'\n');
	add_dbl("pswitch",s->PswitchR,'\n');
	add_dbl("inv_swtch",s->InvSwtchR,'\n');
	add_dbl("cpumigrate",s->MigrateR,'\n');
	add_dbl("xcall",s->XcallsR,'\n');
	add_dbl("lockf",s->LockfR,'\n');
	add_int("Ram",s->Physmem,'\n');
	add_int("Free",s->Pagesfree,'\n');
	add_int("Locked",s->Pageslocked,'\n');
	add_int("SwapTotal",s->SwapTotal,'\n');
	add_int("SwapFree",s->SwapFree,'\n');
	if (s->ZfsArcSize > 0)
		add_int("ZfsArcSize",s->ZfsArcSize,'\n');
	if (s->ScanR >= 0.01)
		add_dbl("scan",s->ScanR,'\n');
	if (s->AnonfreeR >= 0.01)
		add_dbl("anonfree",s->AnonfreeR,'\n');
	if (s->AnonpginR >= 0.01)
		add_dbl("anonpgin",s->AnonpginR,'\n');
	if (s->AnonpgoutR >= 0.01)
		add_dbl("anonpgout",s->AnonpgoutR,'\n');
	/* pgswap, Storage, Fsfree, etc. are sent for backward compability */
	if (s->PgswapinR >= 0.01)
		add_dbl("pgswapin",s->PgswapinR,'\n');
	if (s->PgswapoutR >= 0.01)
		add_dbl("pgswapout",s->PgswapoutR,'\n');
	add_dbl("call",s->SyscallR,'\n');
	if (s->SysexecR >= 0.01)
		add_dbl("exec",s->SysexecR,'\n');
	if (s->SysforkR >= 0.01)
		add_dbl("fork",s->SysforkR,'\n');
	if (s->SysreadR >= 0.01)
		add_dbl("read",s->SysreadR,'\n');
	if (s->SyswriteR >= 0.01)
		add_dbl("write",s->SyswriteR,'\n');
	if (s->IpInR >= 0) {
		add_dbl("IpI",s->IpInR,'\n');
		add_dbl("IpO",s->IpOutR,'\n');
	} else {
		add_dbl("NetI",s->NetInR,'\n');
		add_dbl("NetO",s->NetOutR,'\n');
	}
	add_dbl("NetIb",s->NetInB,'\n');
	add_dbl("NetOb",s->NetOutB,'\n');
	add_int("Storage",s->Storage,'\n');
	add_int("FsFree",s->FsFree,'\n');
	if (s->DiskIR >= 0.1)
		add_dbl("DiskI",s->DiskIR,'\n');
	if (s->DiskOR >= 0.1)
		add_dbl("DiskO",s->DiskOR,'\n');
	if (s->NfsIR >= 0.1)
		add_dbl("NfsI",s->NfsIR,'\n');
	if (s->NfsOR >= 0.1)
		add_dbl("NfsO",s->NfsOR,'\n');
	if (s->NfsdR >= 0.1)
		add_dbl("NfsD",s->NfsdR,'\n');
	if (s->IpInErrs)
		add_int("IpIErr",s->IpInErrs,'\n');
	if (s->IpInDiscards)
		add_int("IpIDis",s->IpInDiscards,'\n');
	if (s->IpOutDiscards)
		add_int("IpODis",s->IpOutDiscards,'\n');
	if (s->HardErrors)
		add_int("HErr",s->HardErrors,'\n');
	if (s->SoftErrors)
		add_int("SErr",s->SoftErrors,'\n');
	while (r) {
		char *m = M;
		msg_t *msg = Msg;

		add_int("cpu",r->id,'\t');
		add_dbl("usr",r->user,'\t');
		add_dbl("sys",r->kernel,'\t');
		add_dbl("idl",r->idle+r->wait,'\t');
		add_dbl("intr",r->intr,'\t');
		add_dbl("ithr",r->intrthread,'\t');
		add_dbl("iblk",r->intrblk,'\t');
		add_dbl("xcal",r->xcalls,'\t');
		add_dbl("migr",r->migrate,'\t');
		add_dbl("icsw",r->inv_swtch,'\t');
		add_dbl("csw",r->pswitch,'\t');
		add_dbl("smtx",r->mxae,'\t');
		add_dbl("rwrf",r->rwrf,'\t');
		add_dbl("rwwf",r->rwwf,'\t');
		add_dbl("asf",r->as_fault,'\t');
		add_dbl("hatf",r->hat_fault,'\t');
		add_dbl("majf",r->majfault,'\t');
		add_dbl("sysc",r->syscall,'\n');
		if (msg != Msg) {
			assert(msg == Msg->next);
			assert(m);
			*m = 0;
			M = Msg->text;
			S = sizeof(Msg->text) - 1;
			add_str("sysstat","1",'\n');
			add_str("Host",SS->Hostinfo,'\n');
		} else
			r = r->next;
	}
	
	while (fs || us || dr) {
		if (us) {
			long long s;
			double pctcpu,pctmem;
			char *m = M;
			msg_t *msg = Msg;
			add_int("uid",us->uid,'\t');
			add_str("user",us->username,'\t');
			add_int("L",us->nlwp,'\t');
			if (us->oncpu)
				add_int("O",us->oncpu,'\t');
			if (us->sleeping)
				add_int("S",us->sleeping,'\t');
			if (us->runnable)
				add_int("R",us->runnable,'\t');
			if (us->zombie)
				add_int("Z",us->zombie,'\t');
			if (us->stopped)
				add_int("T",us->stopped,'\t');
			add_int("SIZE",us->size,'\t');
			add_int("RSS",us->rss,'\t');
			add_int("HS",us->h_s,'\t');
			pctcpu = (double)us->pctcpu/(double)0x8000*100.0;
			pctmem = (double)us->pctmem/(double)0x8000*100.0;
			if (pctmem >= 2E-3)
				add_str("%M",per2str(pctmem),'\t');
			if (pctcpu >= 2E-3)
				add_str("%C",per2str(pctcpu),'\t');
			s = us->time.tv_sec;
			s += (int)rint((double)us->time.tv_nsec/1000000000);
			assert(s >= 0);
			if (s)
				add_int("sec",s,'\t');
			add_int("N",us->nproc,'\n');
			if (msg != Msg) {
				assert(msg == Msg->next);
				assert(m);
				*m = 0;
				M = Msg->text;
				S = sizeof(Msg->text) - 1;
				add_str("sysstat","1",'\n');
				add_str("Host",SS->Hostinfo,'\n');
			} else 
				us = us->next;
		}
		if (fs) {
			char *m = M;
			msg_t *msg = Msg;
			if ((0 == VerboseZFS) && (0 == strcmp(fs->typ,"zfs"))) {
				fs = fs->next;
				continue;
			}
			if (strcmp(fs->typ,"tmpfs")) {
				add_str("dev",fs->dev,'\t');
				add_str("fs",fs->typ,'\t');
				add_str("mnt",fs->mnt,'\t');
			} else {
				add_str("dev","<swap>",'\t');
				add_str("fs","tmpfs",'\t');
				add_str("mnt","<swap>",'\t');
			}
			if (fs->ops_rate >= 0) {
				add_dbl("ops",fs->ops_rate,'\t');
				add_dbl("inr",fs->read_rate,'\t');
				add_dbl("outr",fs->write_rate,'\t');
			}
			add_int("fr",fs->free,'\t');
			if (fs->numzfs)
				add_int("nz",fs->numzfs,'\t');
			add_int("bl",fs->blocks,'\n');
			if (msg != Msg) {
				assert(msg == Msg->next);
				assert(m);
				*m = 0;
				M = Msg->text;
				S = sizeof(Msg->text) - 1;
				add_str("sysstat","1",'\n');
				add_str("Host",SS->Hostinfo,'\n');
			} else 
				fs = fs->next;
		}
		while (dr && (dr->major == -1))
			dr = dr->next;
		if (dr) {
			char *m = M;
			msg_t *msg = Msg;
			if (dr->disk[0])
				add_str("disk",dr->disk,'\t');
			else
				add_str("disk",dr->device,'\t');
			add_dbl("nread",dr->nread,'\t');
			add_dbl("nwritten",dr->nwritten,'\t');
			add_dbl("reads",dr->reads,'\t');
			add_dbl("writes",dr->writes,'\t');
			add_dbl("rbusy",dr->rbusy,'\t');
			add_dbl("wbusy",dr->wbusy,'\n');
			if (msg != Msg) {
				assert(msg == Msg->next);
				assert(m);
				*m = 0;
				M = Msg->text;
				S = sizeof(Msg->text) - 1;
				add_str("sysstat","1",'\n');
				add_str("Host",SS->Hostinfo,'\n');
			} else 
				dr = dr->next;
		}
	}
	while (nr) {
		char *m = M;
		msg_t *msg = Msg;
		add_str("net",nr->device,'\t');
		add_dbl("ipkt",nr->ipckts,'\t');
		add_dbl("opkt",nr->opckts,'\t');
		add_dbl("inb",nr->inbs,'\t');
		add_dbl("outb",nr->outbs,'\n');
		if (msg != Msg) {
			assert(msg == Msg->next);
			assert(m);
			*m = 0;
			M = Msg->text;
			S = sizeof(Msg->text) - 1;
			add_str("sysstat","1",'\n');
			add_str("Host",SS->Hostinfo,'\n');
		} else 
			nr = nr->next;
	}
	*M = 0;
	sendall(localmode);
}
