/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include "sysstat.h"
#include "log.h"
#include "network.h"
#include "terminal.h"

#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#if SOLARIS >= 10
#include <zone.h>
#else
#define ZONENAME_MAX 64
typedef id_t zoneid_t;
zoneid_t (*getzoneid)(void) = 0;
ssize_t (*getzonenamebyid)(zoneid_t,char *,size_t) = 0;
#endif

#ifdef __GNUC__
extern const char *getexecname(void);
#endif

#ifndef PATH_MAX
#define PATH_MAX _XOPEN_PATH_MAX
#endif

int Fslines = 5, Showzfs = 1, Showdisks = 1, Shownets = 1, Localmode = 0, CPULines = 4;
extern int Port;
extern char *McastAddr;

static sysstat_t
	*SS = 0, *Current = 0;
static unsigned
	Curses = 1;

static void
dealloc(sysstat_t *s)
{
	userstat_t *us;
	fsstat_t *fs;
	cpurates_t *r;
	diskrates_t *d;
	
	if (s == 0)
		return;
	us = s->us;
	while (us) {
		userstat_t *n = us->next;
		free(us->username);
		free(us);
		us = n;
	}
	fs = s->fs;
	while (fs) {
		fsstat_t *n = fs->next;
		free(fs->dev);
		free(fs->mnt);
		free(fs);
		fs = n;
	}
	r = s->cpurates;
	while (r) {
		cpurates_t *n = r->next;
		free(r);
		r = n;
	}
	d = s->diskrates;
	while (d) {
		diskrates_t *n = d->next;
		free(d);
		d = n;
	}
	free(s);
}


static void
add_cpurates_by_id(sysstat_t *s, cpurates_t *c)
{
	cpurates_t *x = s->cpurates, *l = 0;

	assert(c);
	if (x == 0) {
		s->cpurates = c;
		c->next = 0;
		return;
	}
	if (c->id < x->id) {
		c->next = x;
		s->cpurates = c;
		return;
	}
	while (x->next && x->next->id < c->id) {
		l = x;
		x = x->next;
	}
	if (x == 0){
		l->next = c;
		c->next = 0;
	} else {
		c->next = x->next;
		x->next = c;
	}
}


static void
add_cpurates_by_idle(sysstat_t *s, cpurates_t *c)
{
	cpurates_t *x = s->cpurates, *l = 0;

	assert(c);
	if (x == 0) {
		s->cpurates = c;
		c->next = 0;
		return;
	}
	if (c->idle < x->idle) {
		c->next = x;
		s->cpurates = c;
		return;
	}
	while (x->next && x->next->idle < c->idle) {
		l = x;
		x = x->next;
	}
	if (x == 0){
		l->next = c;
		c->next = 0;
	} else {
		c->next = x->next;
		x->next = c;
	}
}


static void sort_cpurates(sysstat_t *s)
{
	cpurates_t *cs = s->cpurates;
	s->cpurates = 0;
	while (cs) {
		cpurates_t *nc = cs->next;
		if (s->Ocpus > 4)
			add_cpurates_by_idle(s,cs);
		else
			add_cpurates_by_id(s,cs);
		cs = nc;
	}
}


static void
add_stats(sysstat_t *s)
{
	sysstat_t *i = SS;
	assert((s != i) && (s->next == 0) && (s->prev == 0));
	/* search, if we already have an entry for this host */
	while (i && s) {
		assert(i->next && i->prev);
		if (strcmp(i->Hostinfo,s->Hostinfo)) {
			i = i->next;
			if (i == SS)
				break;
			continue;
		}
		if (i->TS) {
			/* replace this entry */
			dbug("updating stats\n");
			sort_cpurates(s);
			if (i->prev != i) {
				assert(i->next != i);
				s->prev = i->prev;
				assert(s->prev->next == i);
				s->prev->next = s;
				s->next = i->next;
				assert(s->next->prev == i);
				s->next->prev = s;
			} else {
				assert(i->next == i);
				s->prev = s;
				s->next = s;
			}
			if (Current == i)
				Current = s;
			if (SS == i)
				SS = s;
			dealloc(i);
			assert(s->prev && s->next);
			return;
		} else {
			/* add info to this entry */
			/* preserve existing values */
			userstat_t *us = i->us;
			fsstat_t *fs = i->fs;
			cpurates_t *cs = i->cpurates;
			diskrates_t *dr = i->diskrates;
			sysstat_t *next = i->next, *prev = i->prev;
			dbug("appending stats\n");
			memcpy(i,s,sizeof(sysstat_t));
			i->cpurates = cs;
			cs = s->cpurates;
			i->next = next;
			i->prev = prev;
			if (i->fs) {
				fsstat_t *lf = i->fs;
				while (lf->next)
					lf = lf->next;
				lf->next = fs;
			} else
				i->fs = fs;
			if (i->us) {
				userstat_t *lu = i->us;
				while (lu->next)
					lu = lu->next;
				lu->next = us;
			} else
				i->us = us;
			if (i->diskrates) {
				diskrates_t *ld = i->diskrates;
				while (ld->next)
					ld = ld->next;
				ld->next = dr;
			} else
				i->diskrates = dr;
			while (cs) {
				cpurates_t *nc = cs->next;
				if (i->Ocpus > 4)
					add_cpurates_by_idle(i,cs);
				else
					add_cpurates_by_id(i,cs);
				cs = nc;
			}
			free(s);
			return;
		}
	}
	/* new sysstat not already in our list */
	assert(s);
	sort_cpurates(s);
	if (SS) {
		s->next = SS;
		s->prev = SS->prev;
		assert(s->prev->next == SS);
		s->prev->next = s;
		SS->prev = s;
	} else {
		s->next = s;
		s->prev = s;
		SS = s;
		Current = s;
	}
	assert(s->prev && s->next);
}


static void
process_input(int s, int itv)
{
	struct pollfd fds[2];
	char c;
	sysstat_t *st;
	
	fds[0].fd = STDIN_FILENO;
	fds[0].events = POLLIN;
	fds[1].fd = s;
	fds[1].events = POLLIN;
	if (-1 == poll(fds,s == -1 ? 1 : sizeof(fds)/sizeof(struct pollfd),itv * 1000)) {
		if (errno != EINTR)
			error("error polling: %s\n",strerror(errno));
		return;
	}
	if (1 == read(STDIN_FILENO,&c,sizeof(char))) {
		switch (c) {
		case 'h':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->next;
			break;
		case 'H':
			if (Current == 0)
				Current = SS;
			else
				Current = Current->prev;
			break;
		case 'n':
			set_sort_by_nproc();
			if (Current)
				++Current->TS;
			break;
		case 'N':
			Shownets ^= 1;
			if (Current)
				++Current->TS;
			break;
		case 'q':
			exit(EXIT_SUCCESS);
			break;
		case 'r':
			set_sort_by_rss();
			if (Current)
				++Current->TS;
			break;
		case 's':
			set_sort_by_size();
			if (Current)
				++Current->TS;
			break;
		case 't':
			set_sort_by_time();
			if (Current)
				++Current->TS;
			break;
		case 'u':
			set_sort_by_user();
			if (Current)
				++Current->TS;
			break;
		case 'c':
			set_sort_by_pcpu();
			if (Current)
				++Current->TS;
			break;
		case 'C':
			switch (CPULines) {
			default:
				abort();
				break;
			case -1: CPULines = 4; break;
			case 0: CPULines = -1; break;
			case 4: CPULines = 0; break;
			}
			if (Current)
				++Current->TS;
			break;
		case 'd':
			Showdisks ^= 1;
			if (Current)
				++Current->TS;
			break;
		case 'z':
			Showzfs ^= 1;
			if (Current)
				++Current->TS;
			break;
		case '+':
			set_sort_by_hs();
			if (Current)
				++Current->TS;
			break;
		case '?':
			if (Curses) {
				display_help();
				if (Current)
					++Current->TS;
			}
			break;
		default:
			;
		}
	}
	if ((s != -1) && (st = receive_stats(s))) {
		assert(st != SS);
		add_stats(st);
	}
}


static void
usage(const char *m, int r)
{
	printf(
	"sysstat [option] [interval [count]]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-c        : write stats to stdout instead of using curses\n"
	"-d        : start sysstat as daemon, multicasting stats\n"
	"-D        : suppress disk statistics gathering\n"
	"-U        : suppress user statistics gathering\n"
	"-f <num>  : show at least num lines of filesystem information\n"
	"-k        : kill sysstat's daemon\n"
	"-l        : start sysstat in local-mode (i.e. without network support)\n"
	"-m <addr> : send info to spawndaemon address <addr> (default: 224.0.0.75)\n"
	"-p <port> : set spawndaemon port to <port> (default: 7536)\n"
	"-r        : sort user/process summary by rss\n"
	"-s        : sort user/process summary by size\n"
	"-t        : sort user/process summary by %%CPU time\n"
	"-u        : sort user/process summary by username\n"
	"-z        : detailed per zfs statistics\n"
	"-v        : operate in verbose mode\n"
	"-V        : display version number and exit\n\n"
	"%s\n"
	,m
	);
	exit(r);
}


static void
version(void)
{
	printf(
	"sysstat version "VERSION"\n"
	"Copyright 2006-2009, Thomas Maier-Komor\n"
	"This software is licensed under the GPLv3. See file COPYING for details.\n"
	);
	exit(EXIT_SUCCESS);
}


int
main(int argc, char **argv)
{
	unsigned daemon = 0, spawndaemon = 1,
		 iter = 0, killdaemon = 0,
		 nouserstat = 0, nodiskstat = 0;
	int c, count = -1, pid, sock, verbose = 0, fullzfs = 0;
	char *mcastAddr = 0, *port = 0, *itv = 0, *cnt = 0;
	extern char *optarg;
	extern int optind;
	FILE *pidf;
	const char *IntervalArg = 0;

	ARGC = argc;
	ARGV = argv;
	DfltSegV = signal(SIGSEGV,SegVhandler);
#ifdef __sun
	strcpy(Execname,getexecname());
#else
	strcpy(Execname,argv[0]);
#endif
	dbug("Execname %s\n",Execname);
	while ((c = getopt(argc, argv, "cDdf:hklm:p:rstUuvVz")) != -1) {
		switch(c) {
		default:
			assert(0);
			break;
		case 'c':
			Curses = 0;
			break;
		case 'D':
			nodiskstat = 1;
			break;
		case 'd':
			daemon = 1;
			break;
		case 'f':
			if ((optarg == 0) || (1 != sscanf(optarg,"%u",&Fslines)))
				usage("",EXIT_FAILURE);
			break;
		case 'h':
			usage("",EXIT_SUCCESS);
			break;
		case 'k':
			killdaemon = 1;
			break;
		case 'l':
			Localmode = 1;
			break;
		case 'm':
			if (optarg == 0)
				usage("",EXIT_FAILURE);
			McastAddr = optarg;
			mcastAddr = argv[optind-1];
			info("spawndaemon address set to %s\n",optarg);
			break;
		case 'p':
			if ((optarg == 0) || (1 != sscanf(optarg,"%u",&Port)))
				usage("",EXIT_FAILURE);
			info("port set to %s\n",Port);
			port = argv[optind-1];
			break;
		case 'r':
			set_sort_by_rss();
			break;
		case 's':
			set_sort_by_size();
			break;
		case 't':
			set_sort_by_time();
			break;
		case 'U':
			nouserstat = 1;
			break;
		case 'u':
			set_sort_by_user();
			break;
		case 'v':
			verbose = 1;
			setVerbose("sysstat.log");
			break;
		case 'V':
			version();
			break;
		case 'z':
			fullzfs = 1;
			break;
		case ':':
			usage("missing argument\n",EXIT_FAILURE);
			/* FALLTHRU */
		case '?':
			usage("",EXIT_FAILURE);
			break;
		}
	}
	if (optind < argc) {
		unsigned Interval = 1, FSRate = 5, UserRate = 10;
		if (1 > sscanf(argv[optind],"%u/%u/%u",&Interval,&FSRate,&UserRate))
			usage("",EXIT_FAILURE);
		itv = argv[optind];
		IntervalArg = argv[optind];
		dbug("interval set to %u\n",Interval);
		dbug("file system rate set to %u\n",FSRate);
		dbug("user process rate set to %u\n",UserRate);
		++optind;
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%d",&count) || (count <= 0))
			usage("",EXIT_FAILURE);
		cnt = argv[optind];
		++optind;
	}
	if (optind < argc) {
		usage("",EXIT_FAILURE);
	}
	if (-1 == fcntl(STDIN_FILENO,F_SETFL, O_NONBLOCK))
		error("Cannot set input file to non-blocking: %s\n",strerror(errno));
	if ((Localmode == 0) && (pidf = fopen("/tmp/sysstat.pid","r"))) {
		char pn[] = "/proc/\0\0\0\0\0\0",*p;
		struct stat st;
		info("/tmp/sysstat.pid exists\n");
		if (1 != fscanf(pidf,"%d",&pid)) {
			warning("corrupt pid file /tmp/systat.pid - unlinking it\n");
			if (-1 == unlink("/tmp/sysstat.pid"))
				warning("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
		} else {
			p = pn + 6;
			sprintf(p,"%d",pid);
			info("stating %s\n",pn);
			if (-1 == stat(pn,&st)) {
				info("unlinking stale pid file /tmp/sysstat.pid\n");
				if (-1 == unlink("/tmp/sysstat.pid"))
					warning("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
			} else {
				info("daemon already running with pid %d\n",pid);
				spawndaemon = 0;
			}
		}
		fclose(pidf);
	}
	if (killdaemon) {
		char cmd[PATH_MAX];
		int ret = 0;
		strcpy(cmd,dirname(Execname));
		strcat(cmd,"/../../sbin/sysstatd -k");
		ret = system(cmd);
		if (ret == -1) {
			error("could not execute '%s': %s\n",cmd,strerror(errno));
			exit(EXIT_FAILURE);
		}
		exit(ret == 0 ? EXIT_SUCCESS : EXIT_FAILURE);

	}
#ifndef __sun
	spawndaemon = 0;
#endif
	if ((Localmode == 0) && (spawndaemon != 0)) {
		pid_t pid, dpid;
		int err;
#if SOLARIS < 10
		void *libc;
		libc = dlopen("libc.so",RTLD_NOW|RTLD_GLOBAL);
		assert(libc);
		getzoneid = (zoneid_t (*)(void))dlsym(libc,"getzoneid");
		getzonenamebyid = (ssize_t (*)(zoneid_t,char *,size_t))dlsym(libc,"getzonenamebyid");
		if (getzoneid)
#endif
		{
			char zonename[ZONENAME_MAX];
			zoneid_t zone = getzoneid();
			err = getzonenamebyid(zone,zonename,sizeof(zonename));
			assert(err != -1);
			info("running in zone %s\n",zonename);
			if (strcmp(zonename,"global"))
				error("Cannot execute sysstatd in a non-global zone.\n");
		}
		if ((pid = vfork())) {
			int status;
			if (pid == -1)
				error("could not fork: %s\n",strerror(errno));
			waitpid(pid, &status, 0);
			assert(status == 0);
		} else if ((dpid = vfork())) {
			if (dpid == -1)
				error("could not fork: %s\n",strerror(errno));
			exit(EXIT_SUCCESS);
		} else {
			int i;
			char bin[PATH_MAX];
			char *args[16];
			int num = 0;
			strcpy(bin,dirname(Execname));
			strcat(bin,"/../../sbin/sysstatd");
			args[num++] = bin;
			if (verbose)
				args[num++] = "-v";
			if (fullzfs)
				args[num++] = "-z";
			if (port) {
				args[num++] = "-p";
				args[num++] = port;
			}
			if (nodiskstat) 
				args[num++] = "-D";
			if (nouserstat) 
				args[num++] = "-u";
			if (mcastAddr) {
				args[num++] = "-m";
				args[num++] = mcastAddr;
			}
			if (itv)
				args[num++] = itv;
			if (cnt)
				args[num++] = cnt;
			args[num] = 0;
			dbug("starting sysstatd binary %s\n",bin);
			for (i = 0; args[i]; ++i)
				dbug("\targ %d: %s\n",i,args[i]);
			execv(bin,args);
			error("could not exec sysstatd: %s\n",strerror(errno));
		} 
		if (daemon)
			exit(EXIT_SUCCESS);
	} else if (Localmode == 0)
#ifdef __sun
		printf("sysstat daemon already running: waiting for data...\n");
#else
		printf("waiting for data from remote systems...\n");
#endif

	sock = setup_source(IntervalArg,nouserstat,nodiskstat);
	if (Curses)
		Curses = init_curses();
	while ((count == 0) || (iter < count)) {
		process_input(sock,-1);
		if (Curses) {
			iter += display_curses(Current);
		} else {
			iter += display_stdout(Current);
		}
	}
	return 0;
}

// vim:tw=0
