/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <poll.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <unistd.h>
#if SOLARIS >= 10
#include <zone.h>
#else
#define ZONENAME_MAX 64
typedef id_t zoneid_t;
zoneid_t (*getzoneid)(void) = 0;
ssize_t (*getzonenamebyid)(zoneid_t,char *,size_t) = 0;
#endif

#include "log.h"
#include "network.h"
#include "sysstat.h"
#include "terminal.h"
#include "userstat.h"


int Localmode = 0, Userstats = 1, Diskstats = 1, VerboseZFS = 0;
unsigned Interval = 1;


static void
usage(const char *m, int r)
{
	printf(
	"sysstatd [option] [interval [count]]\n"
	"Options:\n"
	"-h        : display this help screen and exit\n"
	"-k        : kill sysstat's daemon\n"
	"-l        : operate in pipe mode\n"
	"-p <port> : set port to send to\n"
	"-d        : do not run as a daemon\n"
	"-D        : disable disk statistics gathering\n"
	"-u        : disable user process statistics gathering\n"
	"-z        : report zfs statistics for all file systems of every pool\n"
	"-v        : operate in verbose mode\n"
	"-V        : display version number and exit\n");
	if (m[0]) 
		printf("\n%s\n",m);
	exit(r);
}


static void
version(void)
{
	printf(
	"sysstat version "VERSION"\n"
	"Copyright 2006-2009, Thomas Maier-Komor\n"
	"This software is licensed under the GPLv3. See file COPYING for details.\n"
	);
	exit(EXIT_SUCCESS);
}


static int
daemon_running(void)
{
	FILE *pidf;
	char pn[] = "/proc/\0\0\0\0\0\0",*p;
	struct stat st;
	int r, pid;

	pidf = fopen("/tmp/sysstat.pid","r");
	if (pidf == 0) 
		return -1;
	info("/tmp/sysstat.pid exists\n");
	r = fscanf(pidf,"%d",&pid);
	fclose(pidf);
	if (1 != r) {
		warning("corrupt pid file /tmp/systat.pid - unlinking it\n");
		if (-1 == unlink("/tmp/sysstat.pid"))
			error("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
		return pid;
	}
	p = pn + 6;
	snprintf(p,6,"%d",pid);
	dbug("stat(\"%s\",...)\n",pn);
	if (-1 == stat(pn,&st)) {
		info("unlinking stale pid file /tmp/sysstat.pid\n");
		if (-1 == unlink("/tmp/sysstat.pid"))
			error("cannot unlink stale pid file /tmp/sysstat.pid: %s\n",strerror(errno));
		return -1;
	}
	return pid;
}


int
main(int argc, char **argv)
{
	unsigned FSRate = 2, UserRate = 10;
	unsigned daemon = 1, killdaemon = 0;
	int c, count = 0, pid;
	extern char *optarg;
	int err;
	zoneid_t zone;
	char zonename[ZONENAME_MAX];

	ARGC = argc;
	ARGV = argv;
#ifdef __sun
	strcpy(Execname,getexecname());
#else
	strcpy(Execname,argv[0]);
#endif
	DfltSegV = signal(SIGSEGV,SegVhandler);
	if (-1 == chdir("/var/tmp"))
		error("Could not change directory to /var/tmp: %s\n",strerror(errno));
	while ((c = getopt(argc, argv, "DdhklmpuvVz")) != -1) {
		extern int Port;
		extern char *McastAddr;

		switch(c) {
		default:
			assert(0);
			break;
		case 'D':
			Diskstats = 0;
			break;
		case 'd':
			daemon = 0;
			setLog(STDERR_FILENO);
			break;
		case 'h':
			usage("",EXIT_SUCCESS);
			break;
		case 'k':
			dbug("kill daemon requested\n");
			killdaemon = 1;
			break;
		case 'l':
			Localmode = 1;
			dbug("local mode requested\n");
			break;
		case 'm':
			McastAddr = optarg;
			info("multicast address set to %s\n",McastAddr);
			break;
		case 'p':
			if ((optarg == 0) || (1 != sscanf(optarg,"%u",&Port)))
				usage("no port given",EXIT_FAILURE);
			break;
		case 'u':
			Userstats = 0;
			dbug("disabling user process statistics lookup");
			break;
		case 'v':
			setVerbose("sysstatd.log");
			break;
		case 'V':
			version();
			break;
		case 'z':
			VerboseZFS = 1;
			break;
		case ':':
			usage("missing argument",EXIT_FAILURE);
			break;
		case '?':
			usage("",EXIT_FAILURE);
			break;
		}
	}
#if SOLARIS < 10
	Libc = dlopen("libc.so",RTLD_NOW|RTLD_GLOBAL);
	assert(Libc);
	getzoneid = (zoneid_t (*)(void))dlsym(Libc,"getzoneid");
	getzonenamebyid = (ssize_t (*)(zoneid_t,char *,size_t))dlsym(Libc,"getzonenamebyid");
	if (getzoneid)
#endif
	{
		zone = getzoneid();
		err = getzonenamebyid(zone,zonename,sizeof(zonename));
		assert(err != -1);
		dbug("running in zone %s\n",zonename);
		if (strcmp(zonename,"global"))
			error("Cannot execute sysstatd in a non-global zone.\n");
	}
	if (optind < argc) {
		if (1 > sscanf(argv[optind],"%u/%u/%u",&Interval,&FSRate,&UserRate))
			usage("",EXIT_FAILURE);
		dbug("interval set to %u\n",Interval);
		dbug("file system rate set to %u\n",FSRate);
		dbug("user process rate set to %u\n",UserRate);
		++optind;
	}
	if (optind < argc) {
		if (1 != sscanf(argv[optind],"%d",&count) || (count <= 0))
			usage("",EXIT_FAILURE);
		dbug("count set to %d\n",count);
		++optind;
	}
	if (optind < argc) {
		usage("",EXIT_FAILURE);
	}
	if (Localmode)
		rundaemon(Interval,FSRate,UserRate,count);
	pid = daemon_running();
	if (-1 != pid) {
		info("daemon already running with pid %d\n",pid);
		if (killdaemon) {
			char aout[PATH_MAX] = "", link[PATH_MAX];
			struct stat st;

			(void)seteuid(0);
			if (-1 == stat("/tmp/sysstat.pid",&st)) {
				error("cannot stat /tmp/sysstat.pid: %s\n",strerror(errno));
				exit(EXIT_FAILURE);
			}
			if (st.st_uid && st.st_uid != getuid()) {
				error("/tmp/sysstat.pid has untrusted uid\n");
				exit(EXIT_FAILURE);
			}
			(void) unlink("/tmp/sysstat.pid");
			sprintf(link,"/proc/%u/path/a.out",pid);
			if (0 == readlink(link,aout,sizeof(aout))) {
				if (strcmp(aout+strlen(aout)-8,"sysstatd") != 0) {
					error("process with pid %u refers to %s instead of sysstatd\n",pid,aout);
					exit(EXIT_FAILURE);
				}
			}
			if (0 == kill(pid,SIGTERM)) {
				info("sending SIGTERM to %d\n",pid);
				exit(EXIT_SUCCESS);
			} else {
				error("cannot send SIGTERM to %d: %s\n",pid,strerror(errno));
				exit(EXIT_FAILURE);
			}
		}
		if (daemon)
			exit(EXIT_FAILURE);
	}
	if (killdaemon) {
		error("cannot kill sysstat daemon: no sysstat daemon running\n");
		exit(EXIT_FAILURE);
	}
	if (daemon != 0) {
		pid_t pid, dpid;
		if ((pid = fork())) {
			int status;
			waitpid(pid, &status, 0);
			assert(status == 0);
		} else if ((dpid = fork()) != 0) {
			FILE *pidf;
			pidf = fopen("/tmp/sysstat.pid","w");
			if (pidf == 0)
				error("cannot create pid file: %s\n",strerror(errno));
			fprintf(pidf,"%d\n",(int)dpid);
			fclose(pidf);
			chmod("/tmp/sysstat.pid",0666);
			exit(EXIT_SUCCESS);
		} else {
			log2syslog(basename(argv[0]));
			rundaemon(Interval,FSRate,UserRate,count);
		} 
		exit(EXIT_SUCCESS);
	} else {
		FILE *pidf;
		if (-1 != daemon_running()) {
			error("sysstatd is already running\n");
			exit(EXIT_FAILURE);
		}
		pidf = fopen("/tmp/sysstat.pid","w");
		if (pidf == 0) {
			error("cannot create pid file: %s\n",strerror(errno));
			exit(EXIT_FAILURE);
		}
		fprintf(pidf,"%d\n",(int)getpid());
		fclose(pidf);
		chmod("/tmp/sysstat.pid",0666);
		rundaemon(Interval,FSRate,UserRate,count);
		if (-1 == unlink("/tmp/sysstat.pid"))
			warning("unable to unlink /tmp/sysstat.pid: %s\n",strerror(errno));
	}
	return 0;
}

// vim:tw=0
