/*
 * Copyright 2006-2009, Thomas Maier-Komor
 * LICENSE: GPLv3, see file COPYING for details
 */

#include "log.h"
#include "sysstat.h"
#include "userstat.h"

#include <assert.h>
#include <dirent.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <procfs.h>
#include <pwd.h>
#include <stdlib.h>
#include <strings.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>


typedef struct vmusage {
	id_t	vmu_zoneid;
	uint_t	vmu_type;
	id_t	vmu_id;
	size_t	vmu_rss_all;
	size_t	vmu_rss_private;
	size_t	vmu_rss_shared;
	size_t	vmu_swap_all;
	size_t	vmu_swap_private;
	size_t	vmu_swap_shared;
} vmusage_t;

typedef struct procentr_s
{
	pid_t pid; /* six chars for pid, one for \0 */
	int psi,pst,upd;
	struct procentr_s *next, *prev;
} procentr_t;

static procentr_t *ProcEntr = 0;
static userstat_t *UserHead = 0;
void *Libc;
static int (*getvmusage)(uint_t flags, time_t age, vmusage_t *buf, size_t *nres);

void init_userstat(void)
{
	if (Libc == 0)
		Libc = dlopen("libc.so",RTLD_NOW|RTLD_GLOBAL);
	assert(Libc);
	getvmusage = (int(*)(uint_t, time_t, vmusage_t *, size_t *)) dlsym(Libc,"getvmusage");
	if (getvmusage)
		dbug("getvmusage is available\n");
}


static userstat_t *
get_userstat(uid_t uid)
{
	userstat_t *u = UserHead;
	while ((u != 0) && (u->uid != uid))
		u = u->next;
	if (u == 0) {
		struct passwd *p;
		u = calloc(1,sizeof(userstat_t));
		assert(u);
		u->uid = uid;
		if ((p = getpwuid(uid))) {
			u->username = strdup(p->pw_name);
		} else {
			size_t s = snprintf(0,0,"uid %d",u->uid) + 1;
			u->username = malloc(s);
			snprintf(u->username,s,"uid %d",u->uid);
		}
		u->next = UserHead;
		UserHead = u;
	}
	return u;
}


static void
add_procentr(procentr_t *e)
{
	procentr_t *c = ProcEntr;
	pid_t pid = e->pid;

	if (c == 0) {
		ProcEntr = e;
		e->next = 0;
		e->prev = 0;
		return;
	}
	if (c->pid < pid) {
		while (c->next && (c->next->pid < pid))
			c = c->next;
		if (c->next == 0) {
			c->next = e;
			e->prev = c;
			e->next = 0;
		} else {
			c->next->prev = e;
			e->next = c->next;
			c->next = e;
			e->prev = c;
		}
	} else {
		assert(pid != c->pid);
		while (c->prev && (c->prev->pid > pid))
			c = c->prev;
		if (c->prev == 0) {
			c->prev = e;
			e->next = c;
			e->prev = 0;
		} else {
			c->prev->next = e;
			e->prev = c->prev;
			c->prev = e;
			e->next = c;
		}
	}
	ProcEntr = e;
}


static procentr_t *
get_procentr(pid_t pid)
{
	procentr_t *c = ProcEntr;
	if (c == 0) {
		return 0;
	}
	if (c->pid == pid) {
		return c;
	} else if (c->pid < pid) {
		do 
			c = c->next;
		while (c && c->pid < pid);
	} else {
		do
			c = c->prev;
		while (c && c->pid > pid);
	}
	if (c && (c->pid == pid)) {
		ProcEntr = c;
		return c;
	}
	return 0;
}


static void
rm_procentr(procentr_t *pe)
{
	if (pe->prev)
		pe->prev->next = pe->next;
	if (pe->next)
		pe->next->prev = pe->prev;
	if (pe == ProcEntr) {
		if (pe->next)
			ProcEntr = pe->next;
		else
			ProcEntr = pe->prev;
	}
	(void) close(pe->psi);
	if (pe->pst != -1)
		(void) close(pe->pst);
	free(pe);
}


static void
cleanup_proc_cache(void)
{
	procentr_t *e = ProcEntr, *p;
	assert(e);
	while (e->next)
		e = e->next;
	do {
		p = e->prev;
		if (e->upd == 0)
			rm_procentr(e);
		else
			e->upd = 0;
		e = p;
	} while (e);
}


static void 
update_procinfo(void)
{
	static int procfd = -1, maxfd, euid;
	static DIR *procdir;
	struct dirent *ent;
	userstat_t *u = UserHead;

	if (-1 == procfd) {
#if SOLARIS >= 9
		procfd = open("/proc", O_RDONLY);
		assert(procfd != -1);
		procdir = fdopendir(procfd);
		assert(procdir);
#else
		procdir = opendir("/proc");
		assert(procdir);
#endif
		maxfd = sysconf(_SC_OPEN_MAX) - 10;
		euid = geteuid();
		if (euid)
			euid = seteuid(0);
	}
	while (u) {
		bzero(u,sizeof(userstat_t)-(sizeof(void *)<<1)-sizeof(uid_t));
/*
		u->nproc = 0;
		u->nlwp = 0;
		u->oncpu = 0;
		u->sleeping = 0;
		u->zombie = 0;
		u->runnable = 0;
		u->stopped = 0;
		u->rss = 0;
		u->size = 0;
		u->h_s = 0;
		u->pctcpu = 0;
		u->pctmem = 0;
		u->time.tv_sec = 0;
		u->time.tv_nsec = 0;
*/
		u = u->next;
	}
	u = UserHead;
	while ((ent = readdir(procdir)) != 0) {
		int psi, pst = -1;
		psinfo_t p;
		pstatus_t s;
		userstat_t *u;
		long long ns;
		pid_t pid;
		procentr_t *pe;

		if (ent->d_name[0] == '.') {
			switch (ent->d_name[1]) {
			case '.': if (ent->d_name[2] != '\0')
					  break;
				  /*FALLTHROUGH*/
			case '\0':
				continue;
			default:;
			}
			/*FALLTHROUGH*/
		}
		/* I think, the above is a little bit more elegant than
		 * this... ;-)
		if (0 == strcmp(".",ent->d_name))
			continue;
		if (0 == strcmp("..",ent->d_name))
			continue;
		*/
		errno = 0;
		pid = strtol(ent->d_name,0,10);
		assert((pid != 0) || (errno != EINVAL));
		pe = get_procentr(pid);
		if (pe != 0) {
			psi = pe->psi;
			pst = pe->pst;
		} else {
#if SOLARIS >= 9
			char psinfo[PATH_MAX], *p = psinfo;
#else
			char psinfo[PATH_MAX] = "/proc/", *p = psinfo+6;
#endif
			size_t dl = strlen(ent->d_name);
			memcpy(p,ent->d_name,dl);
			p += dl;
			memcpy(p,"/psinfo",8);
#if SOLARIS >= 9
			psi = openat(procfd,psinfo,O_RDONLY);
			if (0 == euid) {
				memcpy(p,"/status",8);
				pst = openat(procfd,psinfo,O_RDONLY);
			}
#else
			psi = open(psinfo,O_RDONLY);
			if (0 == euid) {
				memcpy(p,"/status",8);
				pst = open(psinfo,O_RDONLY);
			}
#endif
			if ((-1 == psi) || ((0 == euid) && (pst == -1))) {
				/* process died after reading directory entry */
				if (psi != -1)
					(void) close(psi);
				if (errno == EMFILE)
					error("ran out of file descriptors\n");
				continue;
			}
			pe = malloc(sizeof(procentr_t));
			pe->psi = psi;
			pe->pst = pst;
			pe->pid = pid;
			add_procentr(pe);
		}
		if ((pread(psi,&p,sizeof(p),0) != sizeof(p)) || ((euid == 0) && (pread(pst,&s,sizeof(s),0) != sizeof(s)))) {
			/* process died after opening the file */
			rm_procentr(pe);
			continue;
		}
		if (psi >= maxfd)
			rm_procentr(pe);
		pe->upd = 1;
		u = get_userstat(p.pr_uid);
		++u->nproc;
		u->nlwp += p.pr_nlwp;
		u->size += p.pr_size;
		u->rss += p.pr_rssize;
		if (euid == 0)
			u->h_s += (s.pr_stksize + s.pr_brksize)/1024;
		u->pctcpu += p.pr_pctcpu;
		u->pctmem += p.pr_pctmem;
		u->time.tv_sec += p.pr_time.tv_sec;
		u->time.tv_sec += p.pr_ctime.tv_sec;
		ns = u->time.tv_nsec + p.pr_time.tv_nsec;
		u->time.tv_sec += ns / 1000000000;
		u->time.tv_nsec = ns % 1000000000;
		ns += p.pr_ctime.tv_nsec;
		u->time.tv_sec += ns / 1000000000;
		u->time.tv_nsec = ns % 1000000000;
		switch(p.pr_lwp.pr_sname) {
		default:
			abort();
			break;
		case 'O':
			++u->oncpu;
			break;
		case 'S':
			++u->sleeping;
			break;
		case 'R':
			++u->runnable;
			break;
		case 'Z':
			++u->zombie;
			break;
		case 'T':
			++u->stopped;
			break;
		}
	}
	cleanup_proc_cache();
	rewinddir(procdir);
}


userstat_t *
get_procinfo(void)
{
	update_procinfo();
	if (getvmusage) {
		vmusage_t *v;
		size_t res = 0;
		int ret;
		int c;

		ret = getvmusage(0x1000,5,0,&res);
		assert(ret >= 0);
		v = malloc(res * sizeof(vmusage_t));
		ret = getvmusage(0x1000,5,v,&res);
		assert(ret >= 0);
		for (c = 0; c < res; ++c) {
			userstat_t *u = UserHead;
			while (u && (u->uid != v[c].vmu_id))
				u = u->next;
			if (u) {
				dbug("refining memory information for user %s: rss %d->%d, swap %d->%d\n",
					u->username,
					u->size, v[c].vmu_swap_all >> 10,
					u->rss, v[c].vmu_rss_all >> 10);
				u->size = v[c].vmu_swap_all >> 10;
				u->rss = v[c].vmu_rss_all >> 10;
				u->pctmem = (unsigned short)(((double)(v[c].vmu_rss_all >> 10) / Sysstat.Physmem)*0x8000);
			}
		}
		free(v);
	}
	UserHead = sort_users(UserHead);
	return UserHead;
}

// vim:tw=0
